/*
 * SELM_CouplingOperator_StaggeredGrid1.h
 *
 * http://atzberger.org/
 *
 */

#ifndef SELM_COUPLINGOPERATOR_STAGGEREDGRID1_H_
#define SELM_COUPLINGOPERATOR_STAGGEREDGRID1_H_

#include "SELM_CouplingOperator.h"
#include "SELM_Lagrangian.h"
#include "SELM_Lagrangian_Types.h"
#include "SELM_Lagrangian_CONTROLPTS_BASIC1.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE.h"
#include "SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE.h"
#include "SELM_Eulerian.h"
#include "SELM_Eulerian_Types.h"
#include "SELM_Eulerian_StaggeredGrid1.h"


namespace LAMMPS_NS {

class SELM_CouplingOperator_StaggeredGrid1: public LAMMPS_NS::SELM_CouplingOperator {
//================= function prototypes =================
public:
	SELM_CouplingOperator_StaggeredGrid1();
	SELM_CouplingOperator_StaggeredGrid1(int, char **);
	virtual ~SELM_CouplingOperator_StaggeredGrid1();

	void init();

	void packageError(int errorCode, void *extras);

	void computeOperatorGamma(SELM_Lagrangian *Lpt,
            SELM_Eulerian *Ept, double** u_in, double* u_out);
	void computeOperatorLambda(SELM_Lagrangian *Lpt,
            SELM_Eulerian *Ept, double* f_in, double** f_out);
	void computeOperatorLambda(int numControlPts, double* x,
	            SELM_Eulerian *Ept, double* f_in, double** f_out);
	void computeOperatorUpsilon(SELM_Lagrangian  *Lpt,
			double mu, double meshDeltaX, double* u_in, double* f_out);
	void gradientLambda(int numControlPts, double* x_in,
            double KbT, SELM_Eulerian *Ept, double** f_out);

//  the following function prototypes are not implemented here, but inheriting
//	from base class
	void parse_ParameterFile(char *filename);
    int getOperatorTypeFromStr(char *operatorTypeStr);
    void computeOperator(char            *couplingOpTypeStr,
                         SELM_Lagrangian *SELM_LagrangianData,
                         SELM_Eulerian   *SELM_EulerianData);
    void setup();
    void writeSimulationDataToDisk(char *filename, int timeIndex);

//================= variables =================
	static int          TYPE;
	static char*        TYPE_STR;

	static int          usePeskin;

	int                 numCoupleList;   /* number of lagrangian and eulerian DOF coupled */
	SELM_Lagrangian   **lagrangianList; /* lagrangian info */
	SELM_Eulerian     **eulerianList;   /* eulerian info */

	double              upsilon_0;      /* fluid-particle coupling */


private:
	//Eulerian data
	const static double UNIT_pi;

	const static char* error_str_code;

	double peskinDelta(double r);
	double derivativePeskinDelta(double r);
	double romaDelta(double r);
	double derivativeRomaDelta(double r);
};

}

#endif /* SELM_COUPLINGOPERATOR_STAGGEREDGRID1_H_ */
