/*
 * SELM_CouplingOperator_StaggeredGrid1_XML_Handler.cpp
 *
 * http://atzberger.org/
 *
 */


#include "SELM_CouplingOperator_StaggeredGrid1_XML_Handler.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <malloc.h>

using namespace std;

namespace LAMMPS_NS {

SELM_CouplingOperator_StaggeredGrid1_XML_Handler::SELM_CouplingOperator_StaggeredGrid1_XML_Handler() {
  setupDataHandler();
}

SELM_CouplingOperator_StaggeredGrid1_XML_Handler::SELM_CouplingOperator_StaggeredGrid1_XML_Handler(SELM_CouplingOperator_Delegator_XML_Handler *delegatorHandler) {
  setupDataHandler();

  couplingOp = new SELM_CouplingOperator_StaggeredGrid1();

  strcpy(couplingOp->nameStr, delegatorHandler->SELM_CouplingOperatorName);
  strcpy(couplingOp->typeStr, delegatorHandler->SELM_CouplingOperatorTypeStr);
  strcpy(basePath,            delegatorHandler->basePath);
}

SELM_CouplingOperator_StaggeredGrid1_XML_Handler::SELM_CouplingOperator_StaggeredGrid1_XML_Handler(SELM_CouplingOperator_Delegator_XML_Handler *delegatorHandler,
  int numLagrangianList_in, SELM_Lagrangian **lagrangianList_in,  int numEulerianList_in,  SELM_Eulerian **eulerianList_in) {

  setupDataHandler();

  couplingOp = new SELM_CouplingOperator_StaggeredGrid1();

  strcpy(couplingOp->nameStr, delegatorHandler->SELM_CouplingOperatorName);
  strcpy(couplingOp->typeStr, delegatorHandler->SELM_CouplingOperatorTypeStr);
  strcpy(basePath,            delegatorHandler->basePath);

  numLagrangianList = numLagrangianList_in;
  lagrangianList    = lagrangianList_in;

  numEulerianList   = numEulerianList_in;
  eulerianList      = eulerianList_in;

}

SELM_CouplingOperator_StaggeredGrid1_XML_Handler::~SELM_CouplingOperator_StaggeredGrid1_XML_Handler() {
  delete couplingOp;
}

void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::setupDataHandler() {

  DataHandlerName                           = "Default Data Handler";
  DataHandlerType                           = "SELM_CouplingOperator_StaggeredGrid1_XML_Handler";

  xmlTagName_xml                            = "xml";
  xmlTagName_SELM_CouplingOperator          = "SELM_CouplingOperator";
  xmlTagName_CouplingOperatorName           = "CouplingOperatorName";
  xmlTagName_CouplingOperatorTypeStr        = "CouplingOperatorTypeStr";
  xmlTagName_operatorData                   = "operatorData";

  xmlTagName_numCoupleList                  = "numCoupleList";

  xmlTagName_lagrangianList                 = "lagrangianList";

  xmlTagName_SELM_Lagrangian_Ref            = "SELM_Lagrangian_Ref";

  xmlTagName_LagrangianName                 = "LagrangianName";
  xmlTagName_LagrangianTypeStr              = "LagrangianTypeStr";

  xmlTagName_eulerianList                   = "eulerianList";

  xmlTagName_SELM_Eulerian_Ref              = "SELM_Eulerian_Ref";

  xmlTagName_EulerianName                   = "EulerianName";
  xmlTagName_EulerianTypeStr                = "EulerianTypeStr";

  xmlTagName_T_KERNEL_1_weightTableFilename = "weightTableFilename";
  xmlTagName_Upsilon_0                      = "Upsilon_0";

  xmlTagName_flagWriteSimulationData        = "flagWriteSimulationData";
  xmlTagName_saveSkipSimulationData         = "saveSkipSimulationData";

  couplingOp                   = NULL;
  flagDeterminedCouplingOpType = 0;

  strcpy(basePath,"");

}


void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_startDocument(Atz_XML_SAX_DataHandler* sourceHandler) {

}

void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_endDocument(Atz_XML_SAX_DataHandler* sourceHandler) {

}

void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_startElement(string qName, Atz_XML::AttributesType *attributes, Atz_XML_SAX_DataHandler* sourceHandler) {

  xmlAttributes = attributes;

  xmlString.clear();

  if (qName == xmlTagName_xml) {

  } else if (qName == xmlTagName_SELM_CouplingOperator) {

    /* setup the data structure */
    couplingOp = new SELM_CouplingOperator_StaggeredGrid1();
    flagDeterminedCouplingOpType = 0;

  } else if (qName == xmlTagName_CouplingOperatorName) {

  } else if (qName == xmlTagName_CouplingOperatorTypeStr) {

  } else if (qName == xmlTagName_operatorData) {
//only one operator type
 /*    strcpy(couplingOp->operatorTypeStr, Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
    couplingOp->operatorType     = couplingOp->getOperatorTypeFromStr(couplingOp->operatorTypeStr);
    flagDeterminedCouplingOpType = 1;
*/  } else if (qName == xmlTagName_T_KERNEL_1_weightTableFilename) {

  } else if (qName == xmlTagName_flagWriteSimulationData) {

  } else if (qName == xmlTagName_saveSkipSimulationData) {

  } else if (qName == xmlTagName_numCoupleList) {

  } else if (qName == xmlTagName_lagrangianList) {

  } else if (qName == xmlTagName_SELM_Lagrangian_Ref) {

  } else if (qName == xmlTagName_LagrangianName) {

  } else if (qName == xmlTagName_LagrangianTypeStr) {

  } else if (qName == xmlTagName_eulerianList) {

  } else if (qName == xmlTagName_SELM_Eulerian_Ref) {

  } else if (qName == xmlTagName_EulerianName) {

  } else if (qName == xmlTagName_EulerianTypeStr) {

  } else if (qName == xmlTagName_Upsilon_0) {

  } else { /* unrecognized tags skip (avoid sub-tags triggering something here) */
    Atz_XML_SAX_Handler_Multilevel*    sourceHandler_Multilevel;
    Atz_XML_SAX_DataHandler*           dataHandler;

    sourceHandler_Multilevel = dynamic_cast<Atz_XML_SAX_Handler_Multilevel*>(sourceHandler);
    dataHandler              = new Atz_XML_Helper_Handler_SkipNextTag();
    sourceHandler_Multilevel->parseNextTagWithDataHandler(dataHandler);
  }

}

void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_characters(string xmlString_in, Atz_XML_SAX_DataHandler* sourceHandler) {
  xmlString += xmlString_in;
}

void SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_endElement(string qName, Atz_XML_SAX_DataHandler* sourceHandler) {

  const char *error_str_code = "SELM_CouplingOperators_StaggeredGrid1_XML_Handler.cpp";
  const char *error_str_func = "XML_endElement()";

  if (couplingOp != NULL) {
    /* */
  }

  if (qName == xmlTagName_CouplingOperatorName) {
    strcpy(couplingOp->nameStr, Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
  } else if (qName == xmlTagName_CouplingOperatorTypeStr) {
    strcpy(couplingOp->typeStr, Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
  } else if (qName == xmlTagName_numCoupleList) {

    couplingOp->numCoupleList = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);

    int N = couplingOp->numCoupleList;

    lagrangianNameList      = (char **) malloc(sizeof(char *)*N);
    lagrangianNameList_I    = 0;
    lagrangianTypeStrList   = (char **) malloc(sizeof(char *)*N);
    lagrangianTypeStrList_I = 0;

    eulerianNameList        = (char **) malloc(sizeof(char *)*N);
    eulerianNameList_I      = 0;
    eulerianTypeStrList     = (char **) malloc(sizeof(char *)*N);
    eulerianTypeStrList_I   = 0;

  } else if (qName == xmlTagName_LagrangianName) {
    int I = lagrangianNameList_I;
    int nn = strlen(Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str())+1;
    lagrangianNameList[I] = (char *)malloc(sizeof(char)*nn);
    strcpy(lagrangianNameList[I], Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
    lagrangianNameList_I++;
  } else if (qName == xmlTagName_LagrangianTypeStr) {
    int I = lagrangianTypeStrList_I;
    int nn = strlen(Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str())+1;
    lagrangianTypeStrList[I] = (char *)malloc(sizeof(char)*nn);
    strcpy(lagrangianTypeStrList[I], Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
    lagrangianTypeStrList_I++;
  } else if (qName == xmlTagName_lagrangianList) {

    /* resolve the name references to the master list of lagrangian DOF */
    int N = couplingOp->numCoupleList;
    int M = numLagrangianList;

    couplingOp->lagrangianList = (SELM_Lagrangian **) malloc(sizeof(SELM_Lagrangian *)*N);

    for (int k = 0; k < N; k++) {

      char *nameRef = lagrangianNameList[k];

      couplingOp->lagrangianList[k] = NULL;

      for (int j = 0; j < M; j++) {

        SELM_Lagrangian *lagrangian = lagrangianList[j];
        char            *nameMaster = lagrangian->nameStr;

        if (strcmp(nameRef, nameMaster) == 0) {
          couplingOp->lagrangianList[k] = lagrangian;
        }

      } /* end j loop */

      free(lagrangianNameList[k]);
      free(lagrangianTypeStrList[k]);

    } /* end k loop */

    free(lagrangianNameList);
    free(lagrangianTypeStrList);

  } else if (qName == xmlTagName_EulerianName) {
    int I = eulerianNameList_I;
    int nn = strlen(Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str())+1;
    eulerianNameList[I] = (char *)malloc(sizeof(char)*nn);
    strcpy(eulerianNameList[I], Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
    eulerianNameList_I++;
  } else if (qName == xmlTagName_EulerianTypeStr) {
    int I = eulerianTypeStrList_I;
    int nn = strlen(Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str())+1;
    eulerianTypeStrList[I] = (char *)malloc(sizeof(char)*nn);
    strcpy(eulerianTypeStrList[I], Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
    eulerianTypeStrList_I++;
  } else if (qName == xmlTagName_eulerianList) {

    /* resolve the name references to the master list of eulerian DOF */
    int N = couplingOp->numCoupleList;
    int M = numEulerianList;

    couplingOp->eulerianList = (SELM_Eulerian **) malloc(sizeof(SELM_Eulerian *)*N);

    for (int k = 0; k < N; k++) {

      char *nameRef = eulerianNameList[k];

      couplingOp->eulerianList[k] = NULL;

      for (int j = 0; j < M; j++) {

        SELM_Eulerian *eulerian = eulerianList[j];
        char *nameMaster        = eulerian->nameStr;

        if (strcmp(nameRef,nameMaster) == 0) {
          couplingOp->eulerianList[k] = eulerian;
        }

      } /* end j loop */

      free(eulerianNameList[k]);
      free(eulerianTypeStrList[k]);

    } /* end k loop */

    free(eulerianNameList);
    free(eulerianTypeStrList);

  } else if (qName == xmlTagName_operatorData) {

  } else if (qName == xmlTagName_Upsilon_0) {
    couplingOp->upsilon_0 =  Atz_XML_Helper_ParseData::getDoubleFromAttr(xmlAttributes);

  } else if (flagDeterminedCouplingOpType) {
// only one coupling operator type
  } else { /* unrecognized tags skip (avoid sub-tags triggering something here) */

  }

}

void *SELM_CouplingOperator_StaggeredGrid1_XML_Handler::XML_getData() { /* gets data from parsing the XML */
  return couplingOp;
}

}
