/*
 * SELM_CouplingOperator_Uniform1_Periodic.cpp
 *
 * http://atzberger.org/
 *
 */

#include "SELM_CouplingOperator_Uniform1_Periodic.h"
#include "SELM_CouplingOperator_Types.h"
#include <math.h>
#include <cstdlib>
#include <math.h>
#include "stdio.h"
#include "string.h"

/* LAMMPS includes */
#include "atom.h"
#include "force.h"
#include "update.h"
#include "respa.h"
#include "error.h"
#include "modify.h"
#include "domain.h"

using namespace std;
using namespace LAMMPS_NS;

namespace LAMMPS_NS {

// ============ constants definition ===========
int   SELM_CouplingOperator_Uniform1_Periodic::TYPE     = SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic;
char *SELM_CouplingOperator_Uniform1_Periodic::TYPE_STR = SELM_CouplingOperator_Types::TYPE_STR_Uniform1_Periodic;
const double SELM_CouplingOperator_Uniform1_Periodic::UNIT_pi=3.141592653589793;
const char* SELM_CouplingOperator_Uniform1_Periodic::error_str_code="SELM_CouplingOperator_Uniform1_Periodic.h";

int SELM_CouplingOperator_Uniform1_Periodic::usePeskin  = 1;

SELM_CouplingOperator_Uniform1_Periodic::SELM_CouplingOperator_Uniform1_Periodic() : SELM_CouplingOperator()
{
  /* generic initializations for this class */
  init();
}
SELM_CouplingOperator_Uniform1_Periodic::SELM_CouplingOperator_Uniform1_Periodic(int narg, char **arg) : SELM_CouplingOperator(narg, arg)
{
  const char *error_str_func = "SELM_CouplingOperator_Uniform1_Periodic()";
  /* generic initializations for this class */
  init();
}



SELM_CouplingOperator_Uniform1_Periodic::~SELM_CouplingOperator_Uniform1_Periodic() {
}

void SELM_CouplingOperator_Uniform1_Periodic::packageError(int code, void *extras) {
    exit(code);
}


void SELM_CouplingOperator_Uniform1_Periodic::init() {

  type                       = SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic;
  strcpy(typeStr, SELM_CouplingOperator_Types::TYPE_STR_Uniform1_Periodic);

  upsilon_0                  = 0.0;

  flagWriteSimulationData    = 0;
  saveSkipSimulationData     = 0;

}

void SELM_CouplingOperator_Uniform1_Periodic::computeOperatorGamma(SELM_Lagrangian *Lpt,
                                                                      SELM_Eulerian   *Ept,
                                                                      double** u_in, double* v_out) {

	// multiply Peskin Delta function at position x
	// f_in is the Lagrangian force at position x,
	// f_out is Eulerian force at face center (staggered grid)
    const char* error_str_func="computeOperatorGamma()";

	int k1, k2, k3, i;
	int n, p1, p2, p3, N12, N1;
	double r[3], delta;
	int p[3], i1, i2, i3;
	double L1, L2, L3;

	int    num_dim = 3;
	int    numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];
	double deltaX3;
	int    numMeshPts;

	double *x_ref = NULL;

	int numControlPts;       // Number of control points.

	//stringstream message;
	//message << "Gamma function was found to be broken and needs to be fixed!" << error_str_func << endl;
	//SELM_Package::packageWarning(error_str_code, error_str_func, message);

	// Eulerian data
	if (Ept->type==SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {
      SELM_Eulerian_Uniform1_Periodic* Ept_data=(SELM_Eulerian_Uniform1_Periodic*) Ept;
      num_dim=Ept_data->num_dim;
      for(i=0; i<num_dim; i++) {
        numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
        meshCenterX0[i]=Ept_data->meshCenterX0[i];
      }
      meshDeltaX=Ept_data->meshDeltaX;
	} else {
      stringstream message;
      message << "This method is only implemented on staggered grid" << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

	L1         = numMeshPtsPerDir[0]*meshDeltaX;
	L2         = numMeshPtsPerDir[1]*meshDeltaX;
	L3         = numMeshPtsPerDir[2]*meshDeltaX;

    N1         = numMeshPtsPerDir[0];
    N12        = numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
    deltaX3    = meshDeltaX*meshDeltaX*meshDeltaX;

	// Lagrangian data
	if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
      SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
      numControlPts = Lpt_data->numControlPts;       // Number of control points.
      x_ref         = Lpt_data->ptsX;
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
      SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
      numControlPts = Lpt_data->numControlPts;       // Number of control points.
      x_ref         = Lpt_data->ptsX;
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {
      SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID*) Lpt;
      numControlPts = Lpt_data->numControlPts;       // Number of control points.
      x_ref         = Lpt_data->ptsX;
	} else {
      stringstream message;
      message << "SELM_Lagrangian type is not recognized" << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	// multiply Peskin Delta function at position x
	// u_in is the Eulerian velocity on staggered grid system,
	// v_out is Lagrangian velocity at position x
    double checkSumDelta;
    double checkSumU;

    double xx[3];  // x coordinate in mesh width units
    int    II[3];

    if (usePeskin) {

      for (i = 0; i < numControlPts; i++) {

        // find the grid cell position x lines in, since all the values are face centered,
        // u is shifted by half meshDeltaX on y, z direction
        // v is shifted by half meshDeltaX on x, z direction
        // w is shifted by half meshDeltaX on z, y direction
        i1 = num_dim * i;
        i2 = i1 + 1;
        i3 = i2 + 1;

        II[0] = i1;
        II[1] = i2;
        II[2] = i3;

        xx[0] = (x_ref[i1] - (meshCenterX0[0] - (L1/2.0)))/meshDeltaX;
        xx[1] = (x_ref[i2] - (meshCenterX0[1] - (L2/2.0)))/meshDeltaX;
        xx[2] = (x_ref[i3] - (meshCenterX0[2] - (L3/2.0)))/meshDeltaX;

        // find image in unit cell (make sure xx between [0,N[d]])
        for (int d = 0; d < num_dim; d++) {
          int NN = numMeshPtsPerDir[d];
          if (xx[d] > NN) {
            xx[d] = xx[d] - floor(xx[d]/NN)*NN;
          } else if (xx[d] < 0) {
            xx[d] = xx[d] + ceil(-xx[d]/NN)*NN;
          }
        } /* end d loop */

        v_out[i1] = 0.0;
        v_out[i2] = 0.0;
        v_out[i3] = 0.0;

        for (int comp_d = 0; comp_d < num_dim; comp_d++) {

          // base mesh site index
          p1 = floor(xx[0]);
          p2 = floor(xx[1]);
          p3 = floor(xx[2]);

          checkSumDelta = 0.0;
          checkSumU     = 0.0;
          for (k3 = -2; k3 <= 2; k3++) {
            for (k2 = -2; k2 <= 2; k2++) {
              for (k1 = -2; k1 <= 2; k1++) {

                p[0] = p1 + k1;
                p[1] = p2 + k2;
                p[2] = p3 + k3;

                r[0] = fabs(xx[0] - (p[0] + 0.5));  // note lattice is offset by 0.5 relative to base of mesh
                r[1] = fabs(xx[1] - (p[1] + 0.5));
                r[2] = fabs(xx[2] - (p[2] + 0.5));

                if ((r[0] < 2.0) && (r[1] < 2.0) && (r[2] < 2.0)) {

                  delta = 1.0;
                  for (int d = 0; d < num_dim; d++) {

                    if (p[d] < 0)
                      p[d] += numMeshPtsPerDir[d];

                    if (p[d] >= numMeshPtsPerDir[d])
                      p[d] -= numMeshPtsPerDir[d];

                    if (r[d] <= 1.0) {
                      delta = delta * (3.0 - 2.0*r[d]
                                       + sqrt(1.0 + 4.0*r[d]*(1.0 - r[d]))
                                      ) / 8.0;
                    } else if (r[d] < 2.0) {
                      delta = delta * (5.0 - 2.0*r[d]
                                       - sqrt(-7.0 + 4.0*r[d]*(3.0 - r[d]))
                                      ) / 8.0;
                    } else { // if r > 2.0 then obtain zero value
                      delta = delta * 0.0;
                    }

                  } /* end d loop */

                  n = (p[2]*N12) + (p[1]*N1) + p[0];
                  if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                    stringstream message;
                    message << "Index for the velocity field is out-of-range."
                            << error_str_func << endl;
                    SELM_Package::packageError(error_str_code, error_str_func,
                                               message);
                  }

                  v_out[II[comp_d]] = v_out[II[comp_d]] + (u_in[comp_d][n]*delta); // note factor deltaX3 cancels with integral sum

                  checkSumDelta += delta;
                  checkSumU     += u_in[comp_d][n];
                } /* if r < 2.0 */

              } // end k1
            } // end k2
          } // end k3

          int a = 1;

        } // end comp_d loop

      } // end i loop

    } else { // end of if (usePeskin)
      stringstream message;
      message << "Coupling operator type not specified so nothing computed. " << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }


}

void SELM_CouplingOperator_Uniform1_Periodic::computeOperatorLambda(SELM_Lagrangian  *Lpt,
                                                                       SELM_Eulerian *Ept,
                                                                       double* f_in, double** f_out){

  // multiply Peskin Delta function at position x
  // f_in is the Lagrangian force at position x,
  // f_out is Eulerian force at face center (staggered grid)
  const char* error_str_func = "computeOperatorLambda(SELM_Lagrangia, SELM_Eulerian, f_in, f_out)";

  //Eulerian data
  int     num_dim;
  double  deltaX3;

  int     numMeshPts;

  int     numControlPts; // Number of control points.
  double *x_ref;

  //stringstream message;
  //message << "Lambda function was found to be broken and needs to be debugged!"
  //        << error_str_func << endl;
  //SELM_Package::packageWarning(error_str_code, error_str_func, message);

  //Lagrangian data
  if (Lpt->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
    SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data =
        (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
    numControlPts = Lpt_data->numControlPts; // Number of control points.
    x_ref         = Lpt_data->ptsX;
  } else if (Lpt->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
    SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data =
        (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
    numControlPts = Lpt_data->numControlPts; // Number of control points.
    x_ref         = Lpt_data->ptsX;
  } else if (Lpt->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {
    SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* Lpt_data =
        (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID*) Lpt;
    numControlPts = Lpt_data->numControlPts; // Number of control points.
    x_ref         = Lpt_data->ptsX;
  } else {
    stringstream message;
    message << "SELM_Lagrangian type is not recognized" << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  int flagAdditive = 1; /* atz flag to allow cumulative addition of Lambda operations
                           (assumes user zeros the force array before calling Lambda) */
  computeOperatorLambda(numControlPts, x_ref, Ept, f_in, f_out, flagAdditive);

}

void SELM_CouplingOperator_Uniform1_Periodic::computeOperatorLambda(int numControlPts, double* x_in,
                                                                       SELM_Eulerian *Ept, double* f_in, double** f_out) {

  int flagAdditive = 0; /* atz flag to allow cumulative addition of Lambda operations (set to off by default so field zeroed each time) */

  computeOperatorLambda(numControlPts, x_in, Ept, f_in, f_out, flagAdditive);

}


void SELM_CouplingOperator_Uniform1_Periodic::computeOperatorLambda(int numControlPts, double* x_in,
                                                                       SELM_Eulerian *Ept, double* f_in, double** f_out,
                                                                       int flagAdditive) {

  // int flagAdditive : atz flag to allow cumulative addition of Lambda operations

  // multiply Peskin Delta function at position x
  // f_in is the Lagrangian force at position x,
  // f_out is Eulerian force at face center (staggered grid)
  const char* error_str_func="computeOperatorLambda()";

  int    k1, k2, k3, i;
  int    n, p1, p2, p3, N12, N1, Ntotal;
  double r[3];
  double delta;
  int    p[3], i1, i2, i3;

  double L1, L2, L3;

  //Eulerian data
  int    num_dim = 3;
  int    numMeshPtsPerDir[3];
  double meshDeltaX;
  double meshCenterX0[3];
  double deltaX3;

  int    II[3];
  double xx[3];

  int    numMeshPts;

  if(Ept->type==SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {
      SELM_Eulerian_Uniform1_Periodic* Ept_data=(SELM_Eulerian_Uniform1_Periodic*) Ept;
      num_dim=Ept_data->num_dim;
      for(i=0; i<num_dim; i++) {
        numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
        meshCenterX0[i]=Ept_data->meshCenterX0[i];
      }
      meshDeltaX=Ept_data->meshDeltaX;
  } else {
      stringstream message;
      message << "This method is only implemented on staggered grid" << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  L1 = numMeshPtsPerDir[0]*meshDeltaX;
  L2 = numMeshPtsPerDir[1]*meshDeltaX;
  L3 = numMeshPtsPerDir[2]*meshDeltaX;

  N1        = numMeshPtsPerDir[0];
  N12       = numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
  Ntotal    = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  deltaX3   = meshDeltaX*meshDeltaX*meshDeltaX;

  // flag to allows for additive use of the field f_out
  // if the flag is off (default) we initialize the output array to zero each time
  if (flagAdditive == 0) {
    for (int d = 0; d < num_dim; d++) {
      for (i = 0; i < Ntotal; i++) {
          f_out[d][i] = 0.0;
      }
    }
  } else {
    // leave the f_out alone and allow values to add to the previous results
  }

  if(usePeskin) {

    for(i = 0; i < numControlPts; i++) {

      for (int comp_d = 0; comp_d < num_dim; comp_d++) {

        i1    = num_dim*i;
        i2    = i1 + 1;
        i3    = i2 + 1;

        II[0] = i1;
        II[1] = i2;
        II[2] = i3;

        xx[0] = (x_in[i1] - (meshCenterX0[0] - (L1/2.0)))/meshDeltaX;
        xx[1] = (x_in[i2] - (meshCenterX0[1] - (L2/2.0)))/meshDeltaX;
        xx[2] = (x_in[i3] - (meshCenterX0[2] - (L3/2.0)))/meshDeltaX;

        // find image in unit cell (make sure xx between [0,N[d]])
        for (int d = 0; d < num_dim; d++) {
          int NN = numMeshPtsPerDir[d];
          if (xx[d] > NN) {
            xx[d] = xx[d] - floor(xx[d]/NN)*NN;
          } else if (xx[d] < 0) {
            xx[d] = xx[d] + ceil(-xx[d]/NN)*NN;
          }
        } /* end d loop */

        // f_u component
        p1 = floor(xx[0]);
        p2 = floor(xx[1]);
        p3 = floor(xx[2]);

        double checkSumDelta = 0.0;
        for(k3 = -2; k3 <= 2; k3++) {
          for(k2 = -2; k2 <= 2; k2++) {
            for(k1 = -2; k1 <= 2; k1++) {

              p[0] = p1 + k1;
              p[1] = p2 + k2;
              p[2] = p3 + k3;

              r[0] = fabs(xx[0] - (p[0] + 0.5));
              r[1] = fabs(xx[1] - (p[1] + 0.5));
              r[2] = fabs(xx[2] - (p[2] + 0.5));

              if ((r[0] < 2.0) && (r[1] < 2.0) && (r[2] < 2.0)) {
                delta = 1.0;
                for(int d = 0; d < num_dim; d++) {

                  if (p[d] < 0)
                    p[d] += numMeshPtsPerDir[d];

                  if (p[d] >= numMeshPtsPerDir[d])
                    p[d] -= numMeshPtsPerDir[d];

                  if (r[d] <= 1.0) {
                    delta = delta*(3.0 - 2.0*r[d] + sqrt(1.0 + 4.0*r[d]*(1.0 - r[d])))/8.0;
                  } else if (r[d] < 2.0) {
                    delta = delta*(5.0 - 2.0*r[d] - sqrt(-7.0 + 4.0*r[d]*(3.0 - r[d])))/8.0;
                  } else { // r > 2.0
                    delta = delta*0.0; // contributes zero
                  }

                } // end d loop

                n = (p[2]*N12) + (p[1]*N1) + p[0];
                if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                  stringstream message;
                  message << "Index for the velocity field is out-of-range."
                          << error_str_func << endl;
                  SELM_Package::packageError(error_str_code, error_str_func, message);
                }

                f_out[comp_d][n] = f_out[comp_d][n] + (f_in[II[comp_d]]*delta/deltaX3);
                checkSumDelta   += delta;
              } // end r < 2.0

            } // end k1
          } // end k2
        } // end k3

        int a = 1;

      } // end comp_d
    } // end i

  } // end of if (usePeskin)

}

void SELM_CouplingOperator_Uniform1_Periodic::computeOperatorUpsilon(SELM_Lagrangian  *Lpt,
                                                                        double mu, double meshDeltaX,
                                                                        double* v_in, double* f_out) {
	// Stokes' law:   F_out = 6\pi*mu*R*v
	// Lagrangian description of velocity and force
	// Peskin Delta function radius =2.0r, Roma Delta function radius=1.5r

	int i, i1, i2, i3;
	double c;
	double R;
	int num_dim = 3;
	double numControlPts;

	const char* error_str_func="computeOperatorUpsilon()";

	/*
	if(usePeskin) {
		R=2.0*meshDeltaX;
	} else {
		R=1.5*meshDeltaX;
	}

	c=6.0*UNIT_pi*R*mu;
	*/
	c = upsilon_0;  /* use coupling value upsilon_0 specified when creating the object (XML)*/

	if (upsilon_0 == 0.0) {
	  if ((lammps->update->ntimestep - 1) % 10000 == 0) {
	    stringstream message;
	    message << "Upsilon_0 value is zero.  Perhaps not set in XML." << endl;
	    SELM_Package::packageWarning(error_str_code, error_str_func, message);
	  }
	}

//	c=4.5*mu/(2.0*R*R);
	if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
		SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* Lpt_data=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE*) Lpt;
		numControlPts=Lpt_data->numControlPts;
	} else if(Lpt->type==SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {
		SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* Lpt_data=(SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID*) Lpt;
		numControlPts=Lpt_data->numControlPts;
	} else {
		stringstream message;
		message << "SELM_Lagrangian type is not recognized" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}

	for (i = 0; i < numControlPts; i++) {
		i1        = num_dim*i;
		i2        = i1 + 1;
		i3        = i2 + 1;
		f_out[i1] = c*v_in[i1];
		f_out[i2] = c*v_in[i2];
		f_out[i3] = c*v_in[i3];
	} /* end i loop */
}
void SELM_CouplingOperator_Uniform1_Periodic::writeSimulationDataToDisk(char *baseFilename, int timeIndex) {

  const char *error_str_func = "writeSimulationDataToDisk()";

  FILE *fid;
  char  filename[1000];

  /* open the file for writing the data */
  sprintf(filename, "SELM_CouplingOperator%.9d._%s", timeIndex, typeStr);
  fid = fopen(filename,"w");

  if (fid == NULL) {
    printf("ERROR: %s : %s \n", error_str_code, error_str_func);
    printf("Could not open file, error occured. \n");
    printf("  filename = %s \n", filename);
    packageError(1, 0);
  }

  fprintf(fid, "-- SELM_CouplingOperator_LAMMPS_SHEAR_UNIFORM1_FFTW3_TABLE1 : Simulation Data -- \n");
  fprintf(fid, "\n");

  /* close the file */
  fclose(fid);
}
void SELM_CouplingOperator_Uniform1_Periodic::setup() {

  const char *error_str_func = "setup()";

}

void SELM_CouplingOperator_Uniform1_Periodic::parse_ParameterFile(char *filename) {
  const char* error_str_func="parse_ParameterFile()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

int SELM_CouplingOperator_Uniform1_Periodic::getOperatorTypeFromStr(char *operatorTypeStr) {
  const char* error_str_func="getOperatorTypeFromStr()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

void SELM_CouplingOperator_Uniform1_Periodic::computeOperator(char *couplingOpTypeStr,
  SELM_Lagrangian *SELM_LagrangianData, SELM_Eulerian   *SELM_EulerianData){
  const char* error_str_func="computeOperator()";
  stringstream message;
  message << "This method is not implemented on staggered grid" << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

void SELM_CouplingOperator_Uniform1_Periodic::gradientLambda(int numControlPts, double* x_in,
                                                                double KbT,   SELM_Eulerian *Ept, double** f_out) {
    const char* error_str_func="gradientLambda()";

	int k1, k2, k3, d, i;
	int n, p1, p2, p3, N12, Ntotal;
	double r[3];
	double delta;
	int p[3], i1, i2, i3;
	double meshDeltaX4th;

	double L1, L2, L3;

	//Eulerian data
	int num_dim;
	int numMeshPtsPerDir[3];
	double meshDeltaX;
	double meshCenterX0[3];
	int numMeshPts;

	stringstream message;
	message << "Gradient Lambda function was found to be broken and needs to be fixed!" << error_str_func << endl;
	message << "Not debugged sufficiently yet." << error_str_func << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);

	if(Ept->type==SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {
		SELM_Eulerian_Uniform1_Periodic* Ept_data=(SELM_Eulerian_Uniform1_Periodic*) Ept;
		num_dim=Ept_data->num_dim;
		for(i=0; i<num_dim; i++) {
		  numMeshPtsPerDir[i]=Ept_data->numMeshPtsPerDir[i];
		  meshCenterX0[i]=Ept_data->meshCenterX0[i];
		}
		meshDeltaX=Ept_data->meshDeltaX;
		meshDeltaX4th=meshDeltaX*meshDeltaX*meshDeltaX*meshDeltaX*KbT;
	} else {
		stringstream message;
		message << "This method is only implemented on staggered grid" << endl;
		SELM_Package::packageError(error_str_code, error_str_func, message);
	}


	numMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

	// shift the computational domain to [0,a]x[0,b]x[0,c]
	double* x;
	x=new double[3*numControlPts];

	L1 = numMeshPtsPerDir[0]*meshDeltaX;
	L2 = numMeshPtsPerDir[1]*meshDeltaX;
	L3 = numMeshPtsPerDir[2]*meshDeltaX;

	for(i=0; i<numControlPts; i++) {
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		x[i1]=x_in[i1]-(meshCenterX0[0]-numMeshPtsPerDir[0]*meshDeltaX/2);
		x[i2]=x_in[i2]-(meshCenterX0[1]-numMeshPtsPerDir[1]*meshDeltaX/2);
		x[i3]=x_in[i3]-(meshCenterX0[2]-numMeshPtsPerDir[2]*meshDeltaX/2);

        // find image in unit cell
        if (x[i1] > L1) {
          x[i1] = x[i1] - floor(x[i1]/L1)*L1;
        } else if (x[i1] < 0) {
          x[i1] = x[i1] + ceil(-x[i1]/L1)*L1;
        }

        if (x[i2] > L2) {
          x[i2] = x[i2] - floor(x[i2]/L2)*L2;
        } else if (x[i2] < 0) {
          x[i2] = x[i2] + ceil(-x[i2]/L2)*L2;
        }

        if (x[i3] > L3) {
          x[i3] = x[i3] - floor(x[i3]/L3)*L3;
        } else if (x[i3] < 0) {
          x[i3] = x[i3] + ceil(-x[i3]/L3)*L3;
        }

	}

	N12=numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
	Ntotal=N12*numMeshPtsPerDir[2];

	for(i=0; i<numControlPts; i++) {
		// find the grid cell position x lines in, since all the values are face centered,
		// f_u is shifted by half meshDeltaX on y, z direction
		// f_v is shifted by half meshDeltaX on x, z direction
		// f_w is shifted by half meshDeltaX on z, y direction
		i1=3*i;
		i2=i1+1;
		i3=i2+1;
		// f_u component
		p1=(int)floor(x[i1]/meshDeltaX);
		p2=(int)floor(x[i2]/meshDeltaX-0.5);
		p3=(int)floor(x[i3]/meshDeltaX-0.5);
		if(usePeskin) {
  		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
			  	for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-p[0]);
					r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
					r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[0][n]-=
						derivativePeskinDelta(r[0])*peskinDelta(r[1])*peskinDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }

  		  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
					r[1]=-(x[i2]/meshDeltaX-p[1]);
					r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[1][n]-=
						peskinDelta(r[0])*derivativePeskinDelta(r[1])*peskinDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX);
		  for(k3=-2; k3<3; k3++) {
			 for(k2=-2; k2<3; k2++) {
				for(k1=-2; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
					r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
					r[2]=-(x[i3]/meshDeltaX-p[2]);
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[2][n]-=
						peskinDelta(r[0])*peskinDelta(r[1])*derivativePeskinDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }
		} else {
  		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
			  	for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-p[0]);
					r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
					r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[0][n]-=
						derivativeRomaDelta(r[0])*romaDelta(r[1])*romaDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }

	  	  // f_v component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX);
		  p3=(int)floor(x[i3]/meshDeltaX-0.5);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
					r[1]=-(x[i2]/meshDeltaX-p[1]);
					r[2]=-(x[i3]/meshDeltaX-(p[2]+0.5));
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[1][n]-=
						romaDelta(r[0])*derivativeRomaDelta(r[1])*romaDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }
		  // f_w component
		  p1=(int)floor(x[i1]/meshDeltaX-0.5);
		  p2=(int)floor(x[i2]/meshDeltaX-0.5);
		  p3=(int)floor(x[i3]/meshDeltaX);
		  for(k3=-1; k3<3; k3++) {
			 for(k2=-1; k2<3; k2++) {
				for(k1=-1; k1<3; k1++) {
					p[0]=p1+k1;
					p[1]=p2+k2;
					p[2]=p3+k3;
					r[0]=-(x[i1]/meshDeltaX-(p[0]+0.5));
					r[1]=-(x[i2]/meshDeltaX-(p[1]+0.5));
					r[2]=-(x[i3]/meshDeltaX-p[2]);
					n=p[2]*N12+p[1]*numMeshPtsPerDir[0]+p[0];
                    if ((n < 0) || (n >= numMeshPts)) { /* check fluid velocity index */
                      stringstream message;
                      message << "Index for the velocity field is out-of-range."
                              << error_str_func << endl;
                      SELM_Package::packageError(error_str_code, error_str_func, message);
                    }
					f_out[2][n]-=
							romaDelta(r[0])*romaDelta(r[1])*derivativeRomaDelta(r[2])/meshDeltaX4th;
				}
			 }
		  }
	   }  // end of if(usePeskin)
	}

	delete[] x;
}

double SELM_CouplingOperator_Uniform1_Periodic::peskinDelta(double r){
	if(r>=2||r<=-2) {
		return 0.0;
	} else if (r>-2 && r<=-1) {
		return (5.0+2.0*r-sqrt(-7.0-4.0*r*(3.0+r)))/8.0;
	} else if (r>-1 && r<=0) {
		return (3.0+2.0*r+sqrt(1.0-4.0*r*(1.0+r)))/8.0;
	} else if (r>0 &&  r<=1) {
		return (3.0-2.0*r+sqrt(1.0+4.0*r*(1.0-r)))/8.0;
	} else {
		return (5.0-2.0*r-sqrt(-7.0+4.0*r*(3.0-r)))/8.0;
	}
}

double SELM_CouplingOperator_Uniform1_Periodic::derivativePeskinDelta(double r){
	if(r>=2||r<=-2) {
		return 0.0;
	} else if (r>-2 && r<=-1) {
		return (1.0+(3.0+2.0*r)/sqrt(-7.0-4.0*r*(3.0+r)))/4.0;
	} else if (r>-1 && r<=0) {
		return (1.0-(1.0+2.0*r)/sqrt(1.0-4.0*r*(1.0+r)))/4.0;
	} else if (r>0 &&  r<=1) {
		return (1.0-(-1.0+2.0*r)/sqrt(1.0+4.0*r*(1.0-r)))/4.0;
	} else {
		return (1.0+(-3.0+2.0*r)/sqrt(-7.0+4.0*r*(3.0-r)))/4.0;
	}
}

double SELM_CouplingOperator_Uniform1_Periodic::romaDelta(double r){
	if(r>=1.5 || r<=-1.5) {
		return 0.0;
	} else if(r>-1.5 && r<=-0.5) {
		return (5.0+3.0*r-sqrt(-3.0*(1.0+r)*(1.0+r)+1.0))/6.0;
	} else if(r>-0.5 && r<0.5) {
		return (1.0+sqrt(-3.0*r*r+1.0))/3.0;
	} else {
		return (5.0-3.0*r-sqrt(-3.0*(1.0-r)*(1.0-r)+1.0))/6.0;
	}
}

double SELM_CouplingOperator_Uniform1_Periodic::derivativeRomaDelta(double r){
	if(r>=1.5 || r<=-1.5) {
		return 0.0;
	} else if(r>-1.5 && r<=-0.5) {
		return (1.0+(1.0+r)/sqrt(-3.0*(1.0+r)*(1.0+r)+1.0))/2.0;
	} else if(r>-0.5 && r<0.5) {
		return -r/sqrt(-3.0*r*r+1.0);
	} else {
		return (-1.0+(1.0-r)/sqrt(-3.0*(1.0-r)*(1.0-r)+1.0))/2.0;
	}
}

}
