/*
 * SELM_Eulerian_StaggeredGrid1.cpp
 *
 * http://atzberger.org/
 *
 * WARNING: Used primarily for the Dirichlet codes not Periodic case.
 *
 * WARNING: PJA: made extensive changes not realizing the above, need to replace with old
 * version with Nw and one less in z-direction...
 */

#include "SELM_Eulerian.h"
#include "SELM_Eulerian_Types.h"
#include "SELM_Eulerian_StaggeredGrid1.h"

#include <cstdlib>
#include <iostream>
#include <cstdio>
#include <sstream>
#include <cmath>
#include "string.h"

using namespace std;


namespace LAMMPS_NS {
// ============ constants definition ===========
int SELM_Eulerian_StaggeredGrid1::flag_use_setup    =    0;
double SELM_Eulerian_StaggeredGrid1::UNIT_pi   =    3.141592653589793;
int SELM_Eulerian_StaggeredGrid1::TYPE         =    SELM_Eulerian_Types::TYPE_StaggeredGrid1;
char* SELM_Eulerian_StaggeredGrid1::TYPE_STR  =    SELM_Eulerian_Types::TYPE_STR_StaggeredGrid1;


SELM_Eulerian_StaggeredGrid1::SELM_Eulerian_StaggeredGrid1() {
  // generic initialization
  init();
}
SELM_Eulerian_StaggeredGrid1::SELM_Eulerian_StaggeredGrid1(int narg, char **arg) : SELM_Eulerian(narg, arg)
{

  const char *error_str_code = "SELM_Eulerian_StaggeredGrid1.cpp";
  const char *error_str_func = "SELM_Eulerian_StaggeredGrid1()";

  // generic initialization
  init();
}


SELM_Eulerian_StaggeredGrid1::~SELM_Eulerian_StaggeredGrid1() {
  if(flag_use_setup) {
	  deleteSetup();
  }
}

void SELM_Eulerian_StaggeredGrid1::init() {
	  /* identify the type of the mesh */
	  type = SELM_Eulerian_StaggeredGrid1::TYPE;
	  strcpy(typeStr, SELM_Eulerian_StaggeredGrid1::TYPE_STR);

	  flagWriteSimulationData    = 0;
	  saveSkipSimulationData     = 0;

	  flagWriteFluidVel_VTK      = 0;
	  flagWriteFluidForce_VTK    = 0;
	  flagWriteFluidPressure_VTK = 0;
}

void SELM_Eulerian_StaggeredGrid1::setup() {

  const char *error_str_code = "SELM_Eulerian_StaggeredGrid1.cpp";
  const char *error_str_func = "setup()";

  flag_use_setup = 1;

  //allocate velocity, force density and pressure arrays
  //allocate velocity u, v, w, force  on face center
  //allocate pressure at cell center.
  int N = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  //int Nw=numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*(numMeshPtsPerDir[2]-1);

  for (int d = 0; d < num_dim; d++) {

    u_main[d]         = new double[N];
	u_stochastic[d]           = new double[N];
	u[d]                      = new double[N];
	fluidForceDensity[d]      = new double[N];
	fluidStochForceDensity[d] = new double[N];

	for(int i = 0; i < N; i++) {
	  u_main[d][i]         = 0.0;
	  u_stochastic[d][i]           = 0.0;
	  u[d][i]                      = 0.0;
	  fluidForceDensity[d][i]      = 0.0;
	  fluidStochForceDensity[d][i] = 0.0;
	}

  }

  p_main  = new double[N];
  p_stochastic    = new double[N];
  p               = new double[N];

  for(int i = 0; i < N; i++) {
	p_main[i] = 0.0;
	p_stochastic[i]   = 0.0;
	p[i]              = 0.0;
  }

}

void SELM_Eulerian_StaggeredGrid1::deleteSetup() {

	for (int d=0; d<num_dim; d++) {
	  	delete[] u_main[d];
	  	delete[] u_stochastic[d];
	  	delete[] u[d];
	  	delete[] fluidForceDensity[d];
	}

	delete[] p_main;
	delete[] p_stochastic;
	delete[] p;
}

void SELM_Eulerian_StaggeredGrid1::packageError(int code, void *extras) {
    exit(code);
}

void SELM_Eulerian_StaggeredGrid1::writeSimulationDataToDisk(char *nameStr, int timeIndex) {
	const char *error_str_code = "SELM_Eulerian_StaggeredGrid1.cpp";
	const char *error_str_func = "writeSimulationDataToDisk()";

	int    d, n, k1, k2, k3, N12, I;
	double meshLengths[3];

    char  filename[1000];

    for (d = 0; d < num_dim; d++) {
      meshLengths[d] = numMeshPtsPerDir[d]*meshDeltaX;
    }

    int totalMeshPts;
    totalMeshPts = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
    N12          = numMeshPtsPerDir[0]*numMeshPtsPerDir[1];

    /* @@@ PJA: VTK file writing from staggered mesh is a little unconventional, using averaging
     *  to get cell centered values for mesh??? not sure for z-direction... */

//if(timeIndex%10000==0) {
    if (flagWriteFluidVel_VTK) {

	    sprintf(filename,"%s_FluidVel_%.9d.vtk", nameStr, timeIndex);

	    //interpolate the cell faces velocity to cell center velocity
	    double* u_c[3];

	    for (d = 0; d < num_dim; d++) {
	      u_c[d] = new double[totalMeshPts];
	    }

	    //calculate the kinetic energy
	    double totalKe = 0.0;
	    double rho     = 602.2142*meshDeltaX*meshDeltaX*meshDeltaX;
	    for(k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
	    	for(k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
	    		for(k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

	    			n = k3*N12+k2*numMeshPtsPerDir[0] + k1;

	    			if(k1 == numMeshPtsPerDir[0] - 1) {
	    			  I = k3*N12 + k2*numMeshPtsPerDir[0];
	    			} else {
	    			  I = n + 1;
	    			}
	    			u_c[0][n] = (u_main[0][I] + u_main[0][n])/2.0;

	    			if(k2 == numMeshPtsPerDir[1]-1) {
	    			  I = k3*N12 + k1;
	    			} else {
	    			  I = n + numMeshPtsPerDir[0];
	    			}
	    			u_c[1][n] = (u_main[1][I] + u_main[1][n])/2.0;

	    			if(k3==0) {
	    			  u_c[2][n] = u_main[2][n]/2.0;
	    			} else if(k3 == numMeshPtsPerDir[2] - 1) {
	    			  I         = n - N12;
	    			  u_c[2][n] = u_main[2][I]/2.0;
	    			} else {
	    			  I         = n - N12;
	    			  u_c[2][n] = (u_main[2][I] + u_main[2][n])/2.0;
	    			}
	    			totalKe = rho*(u_c[0][n]*u_c[0][n] + u_c[1][n]*u_c[1][n] + u_c[2][n]*u_c[2][n])/2.0;
	    		}
	    	}
	    }

	    FILE* fid;
	    fid = fopen("LiquidKineticEnergy.dat","a");
	    fprintf(fid,"%g\n", totalKe);
	    fclose(fid);
	    userAppl_writeVecFieldVTKFile( filename, num_dim, numMeshPtsPerDir, meshLengths,"fluid_velocity",u_c);

	    for (d = 0; d < num_dim; d++) {
	      	delete[]  u_c[d];
	    }
	  } // end of flagWriteFluidVel_VTK
//} // end of flag timeIndex%5000==0

	  if (flagWriteFluidForce_VTK) {

	    sprintf(filename,"%s_FluidForce_%.9d.vtk", nameStr, timeIndex);

	    //interpolate the cell faces forces to cell center velocity
	    double* f_c[3];

	    for (d = 0; d < num_dim; d++) {
	      f_c[d]=new double[totalMeshPts];
	    }

	    for(k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
	    	for(k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
	    		for(k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {

	    			n = k3*N12 + k2*numMeshPtsPerDir[0] + k1;

	    			if(k1 == numMeshPtsPerDir[0] - 1) {
	    			  I = k3*N12 + k2*numMeshPtsPerDir[0];
	    			} else {
	    			  I = n + 1;
	    			}

	    			f_c[0][n] = (fluidForceDensity[0][I] + fluidForceDensity[0][n])/2.0;

	    			if(k2 == numMeshPtsPerDir[1] - 1) {
	    			  I = k3*N12 + k1;
	    			} else {
	    			  I = n+numMeshPtsPerDir[0];
	    			}

	    			f_c[1][n] = (fluidForceDensity[1][I] + fluidForceDensity[1][n])/2.0;

	    			if(k3 == 0) {
	    			  f_c[2][n] = fluidForceDensity[2][n]/2.0;
	    			} else if(k3 == numMeshPtsPerDir[2]-1) {
	    			  I         = n - N12;
	    			  f_c[2][n] = fluidForceDensity[2][I]/2.0;
	    			} else {
	    			  I         = n - N12;
	    			  f_c[2][n] = (fluidForceDensity[2][I] + fluidForceDensity[2][n])/2.0;
	    			}

	    		} /* k1 loop */
	    	} /* k2 loop */
	    } /* k3 loop */

	    userAppl_writeVecFieldVTKFile((char *)filename,
	                                       num_dim,
	                                       numMeshPtsPerDir,
	                                       meshLengths,
	                                       "fluid_forceDensity",
	                                       f_c);
	  }// end of flagWriteFluidForce_VTK

	  if (flagWriteFluidPressure_VTK) {
		sprintf(filename,"%s_Pressure_%.9d.vtk", nameStr, timeIndex);

		userAppl_writeScalarFieldVTKFile((char *)filename,
		                                   num_dim,
		                                   numMeshPtsPerDir,
		                                   meshLengths,
		                                   "fluid_Pressure",
		                                   p_main);
	  }


}



void SELM_Eulerian_StaggeredGrid1::userAppl_writeVecFieldVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *vec_name,
                                                                  double **vec_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Vector Field %s. \n",vec_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 0; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5-numMeshPtsPerDir[d]/2.0)*deltaX+meshCenterX0[d];
    	  }

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"VECTORS %s float \n",vec_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  for (d = 0; d < num_dim; d++) {
    		  fprintf(fid,"%g ", vec_array[d][index]);
    	  }
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}

void SELM_Eulerian_StaggeredGrid1::userAppl_writeScalarFieldVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *scalar_name,
                                                                  double *scalar_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Scalar Field %s. \n",scalar_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 0; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5)*deltaX-meshCenterX0[d];
    	  }

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"Scalar_Vectors %s float \n",scalar_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
  		  fprintf(fid,"%g ", scalar_array[index]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");
  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}

void SELM_Eulerian_StaggeredGrid1::userAppl_writeVecFieldXFaceVTKFile(const char   *filename,
                                                                  int     num_dim,
                                                                  int    *numMeshPtsPerDir,
                                                                  double *meshLengths,
                                                                  const char   *vec_name,
                                                                  double **vec_array) {
  FILE  *fid;
  int    k;
  int    d;
  int    J[3];
  double X[3];
  double deltaX;
  int    index;

  int     numIndices;
  int    *indices;

  fid = fopen(filename,"w");

  fprintf(fid,"# vtk DataFile Version 1.0\n");
  fprintf(fid,"Vector Field %s. \n",vec_name);
  fprintf(fid,"ASCII\n");
  fprintf(fid,"\n");

  numIndices = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];
  indices=new int[numIndices];
  for(int k=0; k<numIndices; k++) {
	  indices[k]=k;
  }

  if (numIndices >= 0) {
      fprintf(fid,"DATASET UNSTRUCTURED_GRID\n");
      fprintf(fid,"POINTS %d float\n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  J[2]    = index/(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]);
    	  J[1]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0]))
            		  / numMeshPtsPerDir[0];
    	  J[0]    = (index - J[2]*(numMeshPtsPerDir[1]*numMeshPtsPerDir[0])
    			  - J[1]*numMeshPtsPerDir[0]);

    	  for (d = 1; d < num_dim; d++) {
    		  deltaX = meshLengths[d]/numMeshPtsPerDir[d];
    		  X[d]   = (J[d]+0.5)*deltaX-meshCenterX0[d];
    	  }
    	  X[0]=J[0]*deltaX;

    	  // write the point location
    	  fprintf(fid,"%g %g %g", X[0], X[1], X[2]);
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELLS %d %d\n",numIndices,2*numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 %d \n",k);
      }
      fprintf(fid,"\n");

      fprintf(fid,"CELL_TYPES %d \n",numIndices);
      for (k = 0; k < numIndices; k++) {
    	  fprintf(fid,"1 \n");
      }
      fprintf(fid,"\n");

      fprintf(fid,"POINT_DATA %d \n",numIndices);
      fprintf(fid,"\n");

      fprintf(fid,"VECTORS %s float \n",vec_name);
      for (k = 0; k < numIndices; k++) {
    	  index = indices[k];
    	  for (d = 0; d < num_dim; d++) {
    		  fprintf(fid,"%g ", vec_array[d][index]);
    	  }
    	  fprintf(fid,"\n");
      }
      fprintf(fid,"\n");

  } // end of if numIndices>0
  fclose(fid);
  delete[] indices;
}



}
