/*
 * Stochastic Eulerian Lagrangian Methods
 *
 * Eulerian class storing grid and fluid properties
 *
 * http://atzberger.org/
 *
 */

#ifndef SELM_EULERIAN_STAGGEREDGRID1_H_
#define SELM_EULERIAN_STAGGEREDGRID1_H_

#include "SELM_Eulerian.h"

namespace LAMMPS_NS {

class SELM_Eulerian_StaggeredGrid1: public LAMMPS_NS::SELM_Eulerian {
public:
	SELM_Eulerian_StaggeredGrid1();
	SELM_Eulerian_StaggeredGrid1(int narg, char **arg);
	~SELM_Eulerian_StaggeredGrid1();

	void init();
	void setup();
	void deleteSetup();

	void packageError(int errorCode, void *extras);

    void writeSimulationDataToDisk(char *filename, int timeIndex);


    // ======================== Constants ========================
    static int       flag_use_setup;
    static double    UNIT_pi;
    static int       TYPE;
    static char*     TYPE_STR;

    // variables:
	int 		num_dim;
	int 		numMeshPtsPerDir[3];
	double      meshCenterX0[3];
	double 		meshDeltaX;
    int         flagUseFluidPressure;

	double* 	u_main[3];
	double* 	p_main;
	double* 	u_stochastic[3];
	double* 	p_stochastic;
	double* 	u[3];
	double* 	p;
	double* 	fluidForceDensity[3];
	double*     fluidStochForceDensity[3];

	int         flagWriteFluidVel_VTK;
	int         flagWriteFluidForce_VTK;
	int         flagWriteFluidPressure_VTK;

private:
	void userAppl_writeVecFieldVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *vec_name, double **vec_array);
	void userAppl_writeScalarFieldVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *scalar_name, double *scal_array);
	void userAppl_writeVecFieldXFaceVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *vec_name, double **vec_array);


};

}

#endif /* SELM_EULERIAN_STAGGEREDGRID1_H_ */
