/*
 * SELM_Eulerian_Uniform1_Periodic.h
 *
 * http://atzberger.org/
 *
 */

#ifndef SELM_EULERIAN_UNIFORM1_PERIODIC_H_
#define SELM_EULERIAN_UNIFORM1_PERIODIC_H_

#include "SELM_Eulerian.h"
#include "SELM_Package.h"
#include "Atz_XML_Package.h"

namespace LAMMPS_NS {

class SELM_Eulerian_Uniform1_Periodic: public LAMMPS_NS::SELM_Eulerian {

public:
	SELM_Eulerian_Uniform1_Periodic();
	SELM_Eulerian_Uniform1_Periodic(int narg, char **arg);
	~SELM_Eulerian_Uniform1_Periodic();

	void init();
	void setup();
	void deleteSetup();

	void packageError(int errorCode, void *extras);

	void writeSimulationDataToDisk(char *filename, int timeIndex);
	void setSimulationOutputFlags(char *outputFlagStr);
	void resetSimulationOutputFlags();

	void writeVTKLegacy(char *baseFilename, int timeIndex);
	void writeVTKLegacy(char *filename);

	void writeVTK(char *baseFilename, int timeIndex);
	void writeVTK(char *filename);

	void writeVTK(char *filename, const char *vecFieldName, double **vecField);

	static int writeVTK_vecField(char        *filename,
	                             const char  *vecFieldName,
	                             double     **vecField,
	                             int          num_dim,
	                             int          numMeshPtsPerDir_0,
	                             int          numMeshPtsPerDir_1,
	                             int          numMeshPtsPerDir_2);
	static int writeVTK_vecField(char        *filename,
	                             const char  *vecFieldName,
	                             double     **vecField,
	                             int          num_dim,
	                             double       meshDeltaX,
	                             int          numMeshPtsPerDir_0,
	                             int          numMeshPtsPerDir_1,
	                             int          numMeshPtsPerDir_2,
	                             double       meshCenterX0_0,
	                             double       meshCenterX0_1,
	                             double       meshCenterX0_2);


    // ======================== Constants ========================
    static int         flag_use_setup;
    static double      UNIT_pi;
    static int         TYPE;
    static char*       TYPE_STR;
    static const char *error_str_code;

    // output flags
    const static int       OUTPUTFLAG_TYPE_NULL;
    const static char     *OUTPUTFLAG_TYPESTR_NULL;

    const static int       OUTPUTFLAG_TYPE_SELM;
    const static char     *OUTPUTFLAG_TYPESTR_SELM;

    const static int       OUTPUTFLAG_TYPE_VTK;
    const static char     *OUTPUTFLAG_TYPESTR_VTK;

    const static int       OUTPUTFLAG_TYPE_VTK_LEGACY;
    const static char     *OUTPUTFLAG_TYPESTR_VTK_LEGACY;

    const static int       OUTPUTFLAG_TYPE_VELOCITY;
    const static char     *OUTPUTFLAG_TYPESTR_VELOCITY;

    const static int       OUTPUTFLAG_TYPE_FORCE_DENSITY;
    const static char     *OUTPUTFLAG_TYPESTR_FORCE_DENSITY;

    const static int       OUTPUTFLAG_TYPE_PRESSURE;
    const static char     *OUTPUTFLAG_TYPESTR_PRESSURE;

    const static int       OUTPUTFLAG_TOTAL_NUM;

    // variables:
	int 		num_dim;
	int 		numMeshPtsPerDir[3];
	double      meshCenterX0[3];
	double 		meshDeltaX;
	double      meshL[3];
    int         flagUseFluidPressure;

    // @optimization: should really make these depend on fftw types to avoid
    // later needs for copying into special arrays for performing FFTs.
	double* 	u_main[3];
	double* 	p_main;
	double* 	u_stochastic[3];
	double* 	p_stochastic;
	double* 	u[3];
	double* 	p;
	double* 	fluidForceDensity[3];
	double*     fluidStochForceDensity[3];

	//int         flagWriteFluidVel_VTK;
	//int         flagWriteFluidForce_VTK;
	//int         flagWriteFluidPressure_VTK;

    char           outputFlagsStr[100][100];  // names of output flags
	int            outputFlags[100];          // value of output flag



private:
	void userAppl_writeVecFieldVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *vec_name, double **vec_array);
	void userAppl_writeScalarFieldVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *scalar_name, double *scal_array);
	void userAppl_writeVecFieldXFaceVTKFile(const char *filename, int num_dim, int *numMeshPtsPerDir,
			double *meshLengths, const char *vec_name, double **vec_array);


};

}

#endif /* SELM_EULERIAN_STAGGEREDGRID_PERIOD_H_ */
