/*
 * SELM_Eulerian_Uniform1_Periodic_XML_Handler.cpp
 *
 * http://atzberger.org/
 *
 */

#include "SELM_Eulerian_Uniform1_Periodic_XML_Handler.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

namespace LAMMPS_NS {

char *SELM_Eulerian_Uniform1_Periodic_XML_Handler::error_str_code = "SELM_Eulerian_Uniform1_Periodic_XML_Handler.cpp";

SELM_Eulerian_Uniform1_Periodic_XML_Handler::SELM_Eulerian_Uniform1_Periodic_XML_Handler() {
	 setupDataHandler();
}

SELM_Eulerian_Uniform1_Periodic_XML_Handler::SELM_Eulerian_Uniform1_Periodic_XML_Handler(SELM_Eulerian_Delegator_XML_Handler *delegatorHandler) {

  setupDataHandler();

  // allocate the eulerian object and the parameters data structure
  eulerian = new SELM_Eulerian_Uniform1_Periodic();

  strcpy(eulerian->nameStr, delegatorHandler->SELM_EulerianName);
  strcpy(eulerian->typeStr, delegatorHandler->SELM_EulerianTypeStr);

}

SELM_Eulerian_Uniform1_Periodic_XML_Handler::~SELM_Eulerian_Uniform1_Periodic_XML_Handler() {
	// TODO Auto-generated destructor stub
}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::setupDataHandler() {

  DataHandlerName                         = "Default Data Handler";
  DataHandlerType                         = "SELM_Eulerian_Uniform1_Periodic_XML_Handler";

  xmlTagName_xml                          = "xml";
  xmlTagName_SELM_Eulerian                = "SELM_Eulerian";
  xmlTagName_EulerianName                 = "EulerianName";
  xmlTagName_num_dim                      = "num_dim";
  xmlTagName_numMeshPtsPerDir             = "numMeshPtsPerDir";
  xmlTagName_meshDeltaX                   = "meshDeltaX";
  xmlTagName_meshCenterX0                 = "meshCenterX0";
/*xmlTagName_shearDir                     = "shearDir";
  xmlTagName_shearVelDir                  = "shearVelDir";
  xmlTagName_shearRate                    = "shearRate";
  xmlTagName_shearDist                    = "shearDist"; */
  xmlTagName_flagWriteSimulationData      = "flagWriteSimulationData";
  xmlTagName_saveSkipSimulationData       = "saveSkipSimulationData";
  xmlTagName_outputSimulationData         = "outputSimulationData";

  xmlTagName_flagWriteFluidVel_VTK        = "flagWriteFluidVel_VTK";
  xmlTagName_flagWriteFluidForce_VTK      = "flagWriteFluidForce_VTK";
  xmlTagName_flagWriteFluidPressure_VTK   = "flagWriteFluidPressure_VTK";

  eulerian = NULL;
}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_startDocument(Atz_XML_SAX_DataHandler* sourceHandler) {

}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_endDocument(Atz_XML_SAX_DataHandler* sourceHandler) {

}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_startElement(string qName, Atz_XML::AttributesType *attributes, Atz_XML_SAX_DataHandler* sourceHandler) {

  xmlAttributes = attributes;

  xmlString.clear();

  if (qName == xmlTagName_xml) {

  } else if (qName == xmlTagName_SELM_Eulerian) {

    /* setup the data structure */
    /* allocate the eulerian object and the parameters data structure */
    eulerian = new SELM_Eulerian_Uniform1_Periodic();

  } else if (qName == xmlTagName_EulerianName) {

  } else if (qName == xmlTagName_num_dim) {

  } else if (qName == xmlTagName_numMeshPtsPerDir) {

  } else if (qName == xmlTagName_meshDeltaX) {

  } else if (qName == xmlTagName_meshCenterX0) {

/*  } else if (qName == xmlTagName_shearDir) {

  } else if (qName == xmlTagName_shearVelDir) {

  } else if (qName == xmlTagName_shearRate) {

  } else if (qName == xmlTagName_shearDist) {
*/
  } else if (qName == xmlTagName_flagWriteSimulationData) {

  } else if (qName == xmlTagName_saveSkipSimulationData) {

  } else if (qName == xmlTagName_outputSimulationData) {

  } else if (qName == xmlTagName_flagWriteFluidVel_VTK) {

  } else if (qName == xmlTagName_flagWriteFluidForce_VTK) {

  } else if (qName == xmlTagName_flagWriteFluidPressure_VTK) {


  } else { /* unrecognized tags skip (avoid sub-tags triggering something here) */
    Atz_XML_SAX_Handler_Multilevel*    sourceHandler_Multilevel;
    Atz_XML_SAX_DataHandler*           dataHandler;

    sourceHandler_Multilevel = dynamic_cast<Atz_XML_SAX_Handler_Multilevel*>(sourceHandler);
    dataHandler              = new Atz_XML_Helper_Handler_SkipNextTag();
    sourceHandler_Multilevel->parseNextTagWithDataHandler(dataHandler);
  }

}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_characters(string xmlString_in, Atz_XML_SAX_DataHandler* sourceHandler) {
  xmlString += xmlString_in;
}

void SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_endElement(string qName, Atz_XML_SAX_DataHandler* sourceHandler) {

  const char *error_str_func = "XML_endElement()";

  if (qName == xmlTagName_EulerianName) {
    strcpy(eulerian->nameStr, Atz_XML_Helper_ParseData::getStringFromAttr(xmlAttributes)->c_str());
  } else if (qName == xmlTagName_num_dim) {
    eulerian->num_dim = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_numMeshPtsPerDir) {
    int N             = eulerian->num_dim;                  /* indicates size of pre-alloc array */
    int *intArray_ptr = (int *)&eulerian->numMeshPtsPerDir; /* indicates use pre-alloc array */
    Atz_XML_Helper_ParseData::parseIntArrayFromString(xmlString, &intArray_ptr, &N);
  } else if (qName == xmlTagName_meshDeltaX) {
    stringstream message;
    message << "The meshDeltaX XML tag is no longer supported.  Instead you should simply set" << endl;
    message << "the number of mesh points in each direction to have the same aspect ratio as " << endl;
    message << "the LAMMPS domain.  This might require adjusting the LAMMPS read_data file so " << endl;
    message << "the domain box size has an integer aspect ratio that allows uniform mesh." << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);

    //eulerian->meshDeltaX = Atz_XML_Helper_ParseData::getDoubleFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_meshCenterX0) {
    int     N               = eulerian->num_dim;                 /* indicates size of pre-alloc array */
    double *doubleArray_ptr = (double *)&eulerian->meshCenterX0; /* indicates use pre-alloc array */
    Atz_XML_Helper_ParseData::parseDoubleArrayFromString(xmlString, &doubleArray_ptr , &N);
/*  } else if (qName == xmlTagName_shearDir) {
    eulerianData->shearDir = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_shearVelDir) {
    eulerianData->shearVelDir = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_shearRate) {
    eulerianData->shearRate = Atz_XML_Helper_ParseData::getDoubleFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_shearDist) {
    eulerianData->shearDist = Atz_XML_Helper_ParseData::getDoubleFromAttr(xmlAttributes);
    */
    } else if ((qName == xmlTagName_flagWriteFluidVel_VTK) ||
               (qName == xmlTagName_flagWriteFluidForce_VTK) ||
               (qName == xmlTagName_flagWriteFluidPressure_VTK)) {
    //lagrangian->flagWriteVTK = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
    stringstream message;
    message << "The flagWrite*_VTK XML tags are no longer supported.  Instead you should use the" << endl;
    message << "outputSimulationData tag to specify the data to be written to disk and formats used." << endl;
    message << "xmlTagName_outputSimulationData  = " << xmlTagName_outputSimulationData << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  } else if (qName == xmlTagName_flagWriteSimulationData) {
    eulerian->flagWriteSimulationData = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_saveSkipSimulationData) {
    eulerian->saveSkipSimulationData = Atz_XML_Helper_ParseData::getIntFromAttr(xmlAttributes);
  } else if (qName == xmlTagName_outputSimulationData) {
    eulerian->setSimulationOutputFlags(Atz_XML_Helper_ParseData::getCStringFromAttr(xmlAttributes));
  } else if (qName == xmlTagName_SELM_Eulerian) {
    //eulerian->setup();
  } else { /* unrecognized tags skip (avoid sub-tags triggering something here) */

  }

}

void *SELM_Eulerian_Uniform1_Periodic_XML_Handler::XML_getData() { /* gets data from parsing the XML */
  return eulerian;
}

}
