/*
 * SELM_Integrator_Forward_Euler_Period_Lag.cpp
 *
 * http://atzberger.org/
 * 
 * WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.
 * 
 */


/* SELM includes */
#include "fix_SELM.h"
#include "SELM_Parser1.h"
#include "SELM_Integrator.h"
#include "SELM_Integrator_Forward_Euler_Period_Lag.h"
#include "SELM_Eulerian.h"
#include "SELM_Eulerian_Types.h"
#include "SELM_Eulerian_Uniform1_Periodic.h"
#include "SELM_Lagrangian.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID.h"
#include "SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE.h"
#include "SELM_CouplingOperator_Types.h"
#include "SELM_CouplingOperator_Uniform1_Periodic.h"

/* LAMMPS includes */
#include "atom.h"
#include "force.h"
#include "update.h"
#include "respa.h"
#include "error.h"
#include "modify.h"
#include "domain.h"

/* standard C/C++ includes */
#include <cstdlib>
#include <math.h>
#include "stdio.h"
#include "string.h"
#include <iostream>
#include <sstream>


using namespace std;
using namespace FixConst;


namespace LAMMPS_NS {

const double SELM_Integrator_Forward_Euler_Period_Lag::UNIT_pi  = 3.141592653589793;
const int SELM_Integrator_Forward_Euler_Period_Lag::TYPE        = SELM_Integrator_Types::TYPE_Forward_Euler_Period_Lag;
const char* SELM_Integrator_Forward_Euler_Period_Lag::TYPE_STR  = SELM_Integrator_Types::TYPE_STR_Forward_Euler_Period_Lag;
const char* SELM_Integrator_Forward_Euler_Period_Lag::error_str_code = "SELM_Integrator_Forward_Euler_Period_Lag.cpp";

SELM_Integrator_Forward_Euler_Period_Lag::SELM_Integrator_Forward_Euler_Period_Lag(): SELM_Integrator() {
  printf("WARNING: Creating SELM_Integrator_Forward_Euler_Period_Lag : for this constructor not debugged yet. \n");
  //exit(1);
  init();
}

SELM_Integrator_Forward_Euler_Period_Lag::SELM_Integrator_Forward_Euler_Period_Lag(int narg, char **arg) : SELM_Integrator(narg, arg)
{
  const char *error_str_func = "SELM_Integrator_Forward_Euler_Period_Lag(int narg, char **arg)";
  /* generic initialization */
  init();
  printf("Creating SELM_Integrator_Forward_Euler_Period_Lag: for this constructor not yet implemented. \n");
  exit(1);
}

SELM_Integrator_Forward_Euler_Period_Lag::SELM_Integrator_Forward_Euler_Period_Lag(LAMMPS *lmps, FixSELM *fx){
  const char *error_str_func = "SELM_Integrator_Forward_Euler_Period_Lag(LAMMPS *lmps, Fix *fx)";
  /* generic initialization */
  init();
  setGlobalRefs(lmps, fx);
}

void SELM_Integrator_Forward_Euler_Period_Lag::setGlobalRefs(LAMMPS *lmps, FixSELM *fx) {
  /* setup the LAMMPS related information */
  lammps  = lmps;
  fixSELM = fx;
  /* setup similar to fix_deform.cpp */
  //force_reneighbor = 1;
  //next_reneighbor = -1;
  //if (force_reneighbor)
    irregular = new Irregular(lammps);
  //else
  //  irregular = NULL;
}

void SELM_Integrator_Forward_Euler_Period_Lag::init() {

  type = SELM_Integrator_Forward_Euler_Period_Lag::TYPE;
  strcpy(typeStr, SELM_Integrator_Forward_Euler_Period_Lag::TYPE_STR);

  particleForce_dt      = NULL;
  pseudoForceAll_dt     = NULL;

  lagNumberTimeSteps         = -1;  /* uninitialized */
  lagDeltaT                  = -1.0;
  lagCount                   = 0;

  sum_pseudoForceAll_dt      = NULL;
  sum_pseudoFluidForce_dt[0] = NULL;
  sum_pseudoFluidForce_dt[1] = NULL;
  sum_pseudoFluidForce_dt[2] = NULL;

  totalPts              = 0;
  flagUpdateFluid       = 0; /* only update after lagCount is sufficiently large */

  flagWriteVel = 0;
  if (flagWriteVel) {
    fidVel = fopen("ParticleTest_vel.dat","w");
  } else {
	fidVel = NULL;
  }

  flagUsePureVerlet = 0;  /* 1 - test basic Velocity-Verlet  : 0 - use coupling to fluid */

}

void SELM_Integrator_Forward_Euler_Period_Lag::parse_ParameterFile(char *baseFilename) {

  const char *error_str_func = "parse_ParameterFile()";
}

void SELM_Integrator_Forward_Euler_Period_Lag::setup() {

  const char *error_str_func = "setup()";

  //no longer used here
  //printf("setup() is no longer used in class SELM_Integrator_Forward_Euler_Period_Lag. \n");
  //exit(1);

  /* check that Forward-Euler and update have similar time step */
  if (fabs((deltaT - lammps->update->dt)/deltaT) > 1e-4) {
    stringstream message;
    message << "LAMMPS and SELM time steps do not agree!" << endl;
    message << "LAMMPS deltaT = " << lammps->update->dt << endl;
    message << "SELM deltaT = " << deltaT << endl;
    SELM_Package::packageWarning(error_str_code, error_str_func, message);
  }

  /* setup the lag related data */
  lagDeltaT = lagNumberTimeSteps*deltaT;

}

void SELM_Integrator_Forward_Euler_Period_Lag::packageError(int code, void *extras) {
    exit(code);
}

void SELM_Integrator_Forward_Euler_Period_Lag::set_LAMMPS_mask(int *mask_ptr) {

  /* this determines if LAMMPS calls both the initial and final integrator commands */
  /* this command augments the mask flags to control this behavior */
  (*mask_ptr) |= INITIAL_INTEGRATE; /* tells LAMMPS to call integrate_initial via fix_SELM */
  (*mask_ptr) |= FINAL_INTEGRATE;   /* tells LAMMPS to call integrate_final via fix_SELM */

}

void SELM_Integrator_Forward_Euler_Period_Lag::integrate_initialize() {
  const char *error_str_func = "integrate_initialize()";
}


void SELM_Integrator_Forward_Euler_Period_Lag::integrate_initial() {
  const char *error_str_func = "integrate_initial()";

  /* synchronize the lammps domain with the fluid domain for shear deformation */
  syncDomainWithLammpsDomain();

  /* syncronize the lammps data and control points data to one another */
  for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
    fixSELM->SELM_Lagrangian_List[I]->setControlPtsDataFromLammpsData();
  }

  if (flagUsePureVerlet) {
    integrate_initial_VerletTest1();
  } else {
    // Verlet integration
    SELM_updateFluidAndStructures_initial();
  }

  /* synchronize the lammps data and control points data to one another */
  for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
    fixSELM->SELM_Lagrangian_List[I]->setLammpsDataFromControlPtsData();
  }
  /* synchronize the lammps domain with the fluid domain for shear deformation */
  /* setLammpsDomainFromShearDomain(); */




}

void SELM_Integrator_Forward_Euler_Period_Lag::integrate_final() {

  const char *error_str_func = "integrate_final()";

  /* synchronize the lammps domain with the fluid domain for shear deformation */
  syncDomainWithLammpsDomain();

  /* syncronize the lammps data and control points data to one another */
  for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
    fixSELM->SELM_Lagrangian_List[I]->setControlPtsDataFromLammpsData();
  }

  if (flagUsePureVerlet) {
    integrate_final_VerletTest1();
  } else {
    // perform the calculations before the integrator is called
    start_time_step_Forward_Euler_Period();

    // forward euler integration
    SELM_updateFluidAndStructures_final();
  }

  /* synchronize the lammps data and control points data to one another */
  for (int I = 0; I < fixSELM->SELM_Lagrangian_List_N; I++) {
    fixSELM->SELM_Lagrangian_List[I]->setLammpsDataFromControlPtsData();
  }
  /* synchronize the lammps domain with the fluid domain for shear deformation */
  /* setLammpsDomainFromShearDomain(); */

}


void SELM_Integrator_Forward_Euler_Period_Lag::integrate_initial_VerletTest1() {

  /* Velocity Verlet for the particle degrees of freedom is used

      1) v(t + 0.5*dt) = v(t) + 0.5*a(t)*dt
      2) x(t + dt)     = x(t) + v(t + 0.5*dt)dt
      3) compute a(t + dt) from x(t + dt)
      4) v(t + dt)     = v(t + 0.5*dt) + a(t + dt)*dt
  */

  const char *error_str_func = "integrate_initial_VerletTest1()";

  double  meshDeltaX;
  int     num_dim;
  int    *numMeshPtsPerDir;
  double *meshCenterX0;
  int     numMeshPoints;

  SELM_Eulerian_Uniform1_Periodic *SELM_EulerianData_Uniform1_Periodic = NULL;

  if ((lammps->update->ntimestep - 1) % 1000 == 0) {
    stringstream message;
    message << "Using pure velocity verlet to test ideas." << endl;
    SELM_Package::packageWarning(error_str_code, error_str_func, message);
  }

  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {

      SELM_EulerianData_Uniform1_Periodic = (SELM_Eulerian_Uniform1_Periodic*) fixSELM->SELM_Eulerian_List[0];

      num_dim     = SELM_EulerianData_Uniform1_Periodic->num_dim;
      meshDeltaX  = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;

      numMeshPtsPerDir = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir;
      meshCenterX0     = SELM_EulerianData_Uniform1_Periodic->meshCenterX0;

      numMeshPoints    = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  } else {
    stringstream message;
    message << "Expecting mesh of type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh was used of type: " << fixSELM->SELM_Eulerian_List[0]->typeStr <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* -- update the lagrangian DOF */
  //v(t + 0.5*dt) = v(t) + 0.5*a(t)*dt
  //x(t + dt)     = x(t) + v(t + 0.5*dt)dt
  for (int lagrangianI = 0; lagrangianI < fixSELM->SELM_Lagrangian_List_N; lagrangianI++) {
    SELM_Lagrangian *lagrangian = fixSELM->SELM_Lagrangian_List[lagrangianI];

    if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]*deltaT;

          /* ensure the position is within the periodic domain
           * because if not LAMMPS does not like it! */

          //XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          //if (XX > L) {
          //  XX = XX - floor(XX/L)*L;
          //} else if (XX < 0) {
          //  XX = XX + ceil(-XX/L)*L;
          //}
          //tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]*deltaT;

          /* ensure the position is within the periodic domain
           * because if not LAMMPS does not like it! */

          //XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          //if (XX > L) {
          //  XX = XX - floor(XX/L)*L;
          //} else if (XX < 0) {
          //  XX = XX + ceil(-XX/L)*L;
          //}
          //tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

      SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
        = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]*deltaT;

          /* ensure the position is within the periodic domain
           * because if not LAMMPS does not like it! */

          //XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          //if (XX > L) {
          //  XX = XX - floor(XX/L)*L;
          //} else if (XX < 0) {
          //  XX = XX + ceil(-XX/L)*L;
          //}
          //tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else { /* if not recognized type */
        stringstream message;
      message << "Lagrangian type is not recognized." << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

  } /* end of lagrangianI loop */

}

void SELM_Integrator_Forward_Euler_Period_Lag::integrate_final_VerletTest1() {

  const char *error_str_func = "integrate_final_VerletTest1()";

  /* Velocity Verlet for the particle degrees of freedom is used

      1) v(t + 0.5*dt) = v(t) + 0.5*a(t)*dt
      2) x(t + dt)     = x(t) + v(t + 0.5*dt)dt
      3) compute a(t + dt) from x(t + dt)
      4) v(t + dt)     = v(t + 0.5*dt) + a(t + dt)*dt

      the accelerations are given by

  */

  double  meshDeltaX;
  int     num_dim;
  int    *numMeshPtsPerDir;
  double *meshCenterX0;
  int     numMeshPoints;

  SELM_Eulerian_Uniform1_Periodic *SELM_EulerianData_Uniform1_Periodic = NULL;

  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {

      SELM_EulerianData_Uniform1_Periodic = (SELM_Eulerian_Uniform1_Periodic*) fixSELM->SELM_Eulerian_List[0];

      num_dim     = SELM_EulerianData_Uniform1_Periodic->num_dim;
      meshDeltaX  = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;

      numMeshPtsPerDir = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir;
      meshCenterX0     = SELM_EulerianData_Uniform1_Periodic->meshCenterX0;

      numMeshPoints    = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  } else {
    stringstream message;
    message << "Expecting mesh of type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh was used of type: " << fixSELM->SELM_Eulerian_List[0]->typeStr <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* -- Generate the Langevin forces */

  /* assume operator only one */
  SELM_CouplingOperator_Uniform1_Periodic *op
    = (SELM_CouplingOperator_Uniform1_Periodic *)fixSELM->SELM_CouplingOperator_List[0];

  double upsilon_0 = op->upsilon_0;  /* use value from coupling object (XML) */

  /* -- update the lagrangian DOF */
  //v(t + dt) = v(t + 0.5*dt) + 0.5*a(t + dt)*dt
  for (int lagrangianI = 0; lagrangianI < fixSELM->SELM_Lagrangian_List_N; lagrangianI++) {
    SELM_Lagrangian *lagrangian = fixSELM->SELM_Lagrangian_List[lagrangianI];

    if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          /* drag force using v(t + 0.5*dt) */
          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J] -=
              upsilon_0*SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J];

          /* stochastic force */
          if(flagStochasticDriving) {

            double F_thm;

            int flagRandVariateType = 1;
            if (flagRandVariateType == 1) {  // use Gaussian random variate
              double eta;
              eta   = fixSELM->random->gaussian();
              F_thm = eta*sqrt(2.0*KB*T*upsilon_0/deltaT);
            } else if (flagRandVariateType == 2) { // use Uniform random variate
              double U;
              U     = fixSELM->random->uniform();
              U    -= 0.5; // center the variate
              F_thm = U*sqrt(24.0*KB*T*upsilon_0/deltaT);
            } else {
              F_thm = 0.0;
            }

            double eta;


            /*
             * v(t + dt) = v(t) + 0.5*(a(t) + a(t + dt))*dt
             *
             * m*a(t) = -gamma*v(t - 0.5*dt) + sqrt(2*KBT*gamma/dt)*eta
             *
             * LAMMPS uses uniform random variates and the factor
             *
             * gfactor2 = sqrt(24.0*KB*T*gamma/dt)
             *
             * the 24.0 cancels the centered uniform variate which
             * has variance 1/12.0 to give effective standard deviation
             *
             * sigma = sqrt(2.0*KB*T*gamma/dt).
             *
             */

            SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J] += F_thm;

          } /* end flagStochasticDriving */

          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->atomMass[I];

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          /* drag force using v(t + 0.5*dt) */
          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J] -=
              upsilon_0*SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J];

          /* stochastic force */
          if(flagStochasticDriving) {

            double F_thm;

            int flagRandVariateType = 1;
            if (flagRandVariateType == 1) {  // use Gaussian random variate
              double eta;
              eta   = fixSELM->random->gaussian();
              F_thm = eta*sqrt(2.0*KB*T*upsilon_0/deltaT);
            } else if (flagRandVariateType == 2) { // use Uniform random variate
              double U;
              U     = fixSELM->random->uniform();
              U    -= 0.5; // center the variate
              F_thm = U*sqrt(24.0*KB*T*upsilon_0/deltaT);
            } else {
              F_thm = 0.0;
            }

            double eta;


            /*
             * v(t + dt) = v(t) + 0.5*(a(t) + a(t + dt))*dt
             *
             * m*a(t) = -gamma*v(t - 0.5*dt) + sqrt(2*KBT*gamma/dt)*eta
             *
             * LAMMPS uses uniform random variates and the factor
             *
             * gfactor2 = sqrt(24.0*KB*T*gamma/dt)
             *
             * the 24.0 cancels the centered uniform variate which
             * has variance 1/12.0 to give effective standard deviation
             *
             * sigma = sqrt(2.0*KB*T*gamma/dt).
             *
             */

            SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J] += F_thm;

          } /* end flagStochasticDriving */

          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->atomMass[I];

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

      SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
        = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          /* drag force using v(t + 0.5*dt) */
          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J] -=
              upsilon_0*SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J];

          /* stochastic force */
          if(flagStochasticDriving) {

            double F_thm;

            int flagRandVariateType = 1;
            if (flagRandVariateType == 1) {  // use Gaussian random variate
              double eta;
              eta   = fixSELM->random->gaussian();
              F_thm = eta*sqrt(2.0*KB*T*upsilon_0/deltaT);
            } else if (flagRandVariateType == 2) { // use Uniform random variate
              double U;
              U     = fixSELM->random->uniform();
              U    -= 0.5; // center the variate
              F_thm = U*sqrt(24.0*KB*T*upsilon_0/deltaT);
            } else {
              F_thm = 0.0;
            }

            double eta;


            /*
             * v(t + dt) = v(t) + 0.5*(a(t) + a(t + dt))*dt
             *
             * m*a(t) = -gamma*v(t - 0.5*dt) + sqrt(2*KBT*gamma/dt)*eta
             *
             * LAMMPS uses uniform random variates and the factor
             *
             * gfactor2 = sqrt(24.0*KB*T*gamma/dt)
             *
             * the 24.0 cancels the centered uniform variate which
             * has variance 1/12.0 to give effective standard deviation
             *
             * sigma = sqrt(2.0*KB*T*gamma/dt).
             *
             */

            SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J] += F_thm;

          } /* end flagStochasticDriving */

          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->atomMass[I];

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */


    } else { /* if not recognized type */
        stringstream message;
      message << "Lagrangian type is not recognized." << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

  } /* end of lagrangianI loop */

}



void SELM_Integrator_Forward_Euler_Period_Lag::SELM_updateParticlesTest_initial()
{
  const char *error_str_func = "SELM_updateParticlesTest_initial()";

  /* do nothing */
}

void SELM_Integrator_Forward_Euler_Period_Lag::SELM_updateParticlesTest_final()
{

  Atom   *atom    = lammps->atom;
  int    igroup   = fixSELM->igroup;
  int    groupbit = fixSELM->groupbit;

  double **x      = atom->x;
  double **v      = atom->v;
  double **f      = atom->f;
  double *rmass   = atom->rmass;
  double *mass    = atom->mass;
  int    *type    = atom->type;
  int    *mask    = atom->mask;
  int     nlocal  = atom->nlocal;

}

void SELM_Integrator_Forward_Euler_Period_Lag::SELM_updateFluidAndStructures_final()
{

  const char *error_str_func ="SELM_updateFluidAndStructures_final()";

  if (particleForce_dt == NULL) {
    stringstream message;
    message << "start_time_step_Forward_Euler_Period() must be called before:"
            << error_str_func << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  SELM_Eulerian_Uniform1_Periodic* SELM_EulerianData_Uniform1_Periodic
    = (SELM_Eulerian_Uniform1_Periodic*) fixSELM->SELM_Eulerian_List[0];

  int     num_dim = SELM_EulerianData_Uniform1_Periodic->num_dim; // expect three dimension only

  int 	  numMeshPtsPerDir[3];
  double  meshCenterX0[3];
  double  meshDeltaX;
  int     flagUseFluidPressure;

  double* u_main[3];
  double* p_main;
  double* u_stochastic[3];
  double* p_stochastic;
  double* u[3];
  double* p;

  int     I;
  int     numMeshPts       = -1;
  int     numMeshPtsOnlyXY = -1;
  double  dxsq, lap_u, lap_v, lap_w;
  double  lo[3], hi[3];  // computational domain

  int                          d;
  int                          k1, k2, k3;

  // -- Collect information and dereference pointers
  meshDeltaX     = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;

  p_main = SELM_EulerianData_Uniform1_Periodic->p_main;
  p_stochastic   = SELM_EulerianData_Uniform1_Periodic->p_stochastic;

  p              = SELM_EulerianData_Uniform1_Periodic->p;

  for(d = 0; d < num_dim; d++) {
    meshCenterX0[d]     = SELM_EulerianData_Uniform1_Periodic->meshCenterX0[d];
    numMeshPtsPerDir[d] = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir[d];

    u_main[d]   = SELM_EulerianData_Uniform1_Periodic->u_main[d];
    u_stochastic[d]     = SELM_EulerianData_Uniform1_Periodic->u_stochastic[d];
    u[d]                = SELM_EulerianData_Uniform1_Periodic->u[d];

    lo[d]               = meshCenterX0[d] - (meshDeltaX*numMeshPtsPerDir[d]/2.0);
    hi[d]               = meshCenterX0[d] + (meshDeltaX*numMeshPtsPerDir[d]/2.0);
  } /* end d loop */

  numMeshPtsOnlyXY = numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
  numMeshPts       = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  /* determine if we actually update the fluid this timestep or lag */
  /*
  ==============================================================================
  if (flagUpdateFluid)
  ==============================================================================
  */
  if (flagUpdateFluid) {

    /* WARNING!!!! need to implement the fluid update step in a manner consistent with
     * the Velocity-Verlet scheme to ensure conservation of total momentum of the system.
     *
     * This can be readily done, but need to modify the scheme below!!!!!
     *
     */

    SELM_Integrator_FFTW3_Period fftwop;

    double*                      u_temp[3];
    double*                      u_laplacian[3];
    double*                      p_temp;

    // -- Allocate memory for temporary arrays
    p_temp           = new double[numMeshPts];

    u_temp[0]        = new double[numMeshPts];
    u_temp[1]        = new double[numMeshPts];
    u_temp[2]        = new double[numMeshPts];

    // -- Laplacian Step
    /*
    for(d=0; d<num_dim; d++) {
        lap_u[d]=new double[N];
    }

    SELM_Integrator_FFTW3_Period fftwop;
    */

    //compute laplancian of u
    /*
    fftwop.compute_Laplacian(meshDeltaX, numMeshPtsPerDir, u_main, lap_u);
    */

    // update the fluid velocity to next time step.
    /*
    for(d=0; d<num_dim; d++) {
      for(I=0; I<N; I++) {
        u_main[d][I]+=(mu*lap_u[d][I]*deltaT+fluidForceDensity_lagDt[d][I])/rho;
      }
    }
    */

    // -- Laplacian Step
    for(d = 0; d < num_dim; d++) {
      u_laplacian[d] = u_temp[d];
    }
    //fftwop.compute_Laplacian(meshDeltaX, numMeshPtsPerDir, u_main, u_laplacian);
    computeLaplacian(meshDeltaX, numMeshPtsPerDir, u_main, u_laplacian);

    // -- Update the fluid state
    int N1 = numMeshPtsPerDir[0];
    int N2 = numMeshPtsPerDir[1];
    int N3 = numMeshPtsPerDir[2];
    for (k3 = 0; k3 < N3; k3++) {
      for (k2 = 0; k2 < N2; k2++) {
        for (k1 = 0; k1 < N1; k1++) {

          I = k3*N1*N2 + k2*N1 + k1;

          for(d = 0; d < num_dim; d++) {
            u_main[d][I] += ((mu*u_laplacian[d][I]*lagDeltaT)
                                  + fluidForceDensity_lagDt[d][I])/rho;
          }

        }
      }
    }

    // -- Projection Step
    // compute divergence and inverse Laplacian
    fftwop.compute_Divergence(meshDeltaX, numMeshPtsPerDir,
                              u_main, p_temp);

    fftwop.inverseLaplacian1D(meshDeltaX, numMeshPtsPerDir,
                              p_temp, p_main);

    // determine pressure
    for (I = 0; I < numMeshPts; I++) {
      p_temp[I]         = p_main[I];
      p_main[I] = p_main[I] * rho / deltaT;
    }

    // compute gradient correction of fluid
    fftwop.compute_Gradient(meshDeltaX, numMeshPtsPerDir, p_temp, u_temp);
    fftwop.vel_Diff(numMeshPtsPerDir, u_main, u_temp);

    // -- Free memory from temporary arrays
    delete[] p_temp;
    p_temp = NULL;
    for(I = 0; I < num_dim; I++){
      delete[] u_temp[I];
      u_temp[I] = NULL;
      delete[] fluidForceDensity_lagDt[I];
      fluidForceDensity_lagDt[I] = NULL;
    }

    // -- Set fluid to zero
    int flagSetFluidZero = 0;
    if (flagSetFluidZero) {

      if (((lammps->update->ntimestep - 1) % 1000 == 0) || (lammps->update->ntimestep < 10)) {
        stringstream message;
        message << "Manually setting fluid velocity to zero for debug test." << endl;
        SELM_Package::packageWarning(error_str_code, error_str_func, message);
      }

      int N1 = numMeshPtsPerDir[0];
      int N2 = numMeshPtsPerDir[1];
      int N3 = numMeshPtsPerDir[2];

      for (k3 = 0; k3 < N3; k3++) {
        for (k2 = 0; k2 < N2; k2++) {
          for (k1 = 0; k1 < N1; k1++) {

            I = k3*N1*N2 + k2*N1 + k1;

            u_main[0][I] = 0.0;
            u_main[1][I] = 0.0;
            u_main[2][I] = 0.0;

          }
        }
      }

    } /* end flagSetFluidZero */


    // reset the lag count and fluid updating for the next time step
    lagCount        = 0;
    flagUpdateFluid = 0; // do not update fluid again until lag count sufficient

  } /* end flagUpdateFluid */
  /*
  ==============================================================================
  ==============================================================================
  */

  /* loop over the coupling operators DOF */
  int flagNewWayLagrangianUpdate = 1;

  if (flagNewWayLagrangianUpdate == 1) {

    /* -- add the pseudo-force (drag force) to the particles
     *    use same indexing as in coupling operator calculations via same loops */
	  for (int couplingOpI = 0; couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {

	    SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

	    if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic) {

        SELM_CouplingOperator_Uniform1_Periodic *op
          = (SELM_CouplingOperator_Uniform1_Periodic *)couplingOp;

        /* loop over the coupled entities */
        for (int k = 0; k < op->numCoupleList; k++) {
          SELM_Lagrangian *lagrangian = op->lagrangianList[k];

          /* == Dereference the data */
          if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

            SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
              = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

            int numDOFControlPts = SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts*num_dim;

            for(int i = 0; i < numDOFControlPts; i++) {
              SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[i] += particleForce_dt[couplingOpI][k][i]/deltaT;  /* note dt was included to handle thermal forces, so remove */
              /* NOTE: LAMMPS forces do not have deltaT included in them, so need to remove. need to (@optimize) */
            } /* end i loop */

          } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

            SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
              = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

            int numDOFControlPts = SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts*num_dim;

            for(int i = 0; i < numDOFControlPts; i++) {
              SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[i] += particleForce_dt[couplingOpI][k][i]/deltaT;  /* note dt was included to handle thermal forces, so remove */
              /* NOTE: LAMMPS forces do not have deltaT included in them, so need to remove. need to (@optimize) */
            } /* end i loop */

          } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

            SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
              = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

            int numDOFControlPts = SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numControlPts*num_dim;

            for(int i = 0; i < numDOFControlPts; i++) {
              SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[i] += particleForce_dt[couplingOpI][k][i]/deltaT;  /* note dt was included to handle thermal forces, so remove */
              /* NOTE: LAMMPS forces do not have deltaT included in them, so need to remove. need to (@optimize) */
            } /* end i loop */

          } else { /* if not recognized type */
            stringstream message;
            message << "Lagrangian type is not recognized." << endl;
            SELM_Package::packageError(error_str_code, error_str_func, message);
          }

          delete[] particleForce_dt[couplingOpI][k];
        } /* end k loop */

      } /* end if coupling type = ZZZ */

      delete[] particleForce_dt[couplingOpI];

	  } /* end couplingOpI loop */

	  delete[] particleForce_dt;

	  particleForce_dt = NULL;  /* (@optimize) allocate and deallocate only once or as needed */

	  /* -- update the lagrangian DOF */
    for (int lagrangianI = 0; lagrangianI < fixSELM->SELM_Lagrangian_List_N; lagrangianI++) {
      SELM_Lagrangian *lagrangian = fixSELM->SELM_Lagrangian_List[lagrangianI];

      if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

        SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
          = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

        for (I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; I++) {
          for (int d = 0; d < num_dim; d++) {

            int J = (I*num_dim) + d;

            /* position x(t + dt) = x(t) + v(t + 0.5*dt)*dt was already updated in initial
             * part of the integration step.
             *
             * The main task here is update the fluid velocity the rest of the half-step.
             *
             * WARNING: Need to make sure above Gamma and Lambda operators are evaluated
             * at x(t) and not x(t + dt) if we want to achieve the Forward-Euler scheme
             * for the fluid equations.  Otherwise may get funny behaviors?
             *
             * WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.
             *
             */
            /*
            double tmpPtsX;
            double XX;
            double L = meshDeltaX*numMeshPtsPerDir[d];
            tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J]
                    + SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]*deltaT;

            // ensure the position is within the periodic domain
            // because if not LAMMPS does not like it!
            XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
            if (XX > L) {
              XX = XX - floor(XX/L)*L;
            } else if (XX < 0) {
              XX = XX + ceil(-XX/L)*L;
            }
            tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

            SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J] = tmpPtsX;
            */

            // v(t + dt) = v(t) + 0.5*a(t + dt)*dt
            SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]
              += (SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
               / SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->atomMass[I];

           /* NOTE: LAMMPS forces do not have deltaT included in them. */

           if (flagWriteVel) {
             if (d == 0) {
               fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]);
               fflush(fidVel);
             }
           }

          } /* end of d loop */
        } /* end I loop */

      } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

        SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
          = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

        for (I = 0; I < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; I++) {
          for (int d = 0; d < num_dim; d++) {

            int J = (I*num_dim) + d;

            /* position x(t + dt) = x(t) + v(t + 0.5*dt)*dt was already updated in initial
             * part of the integration step.
             *
             * The main task here is update the fluid velocity the rest of the half-step.
             *
             * WARNING: Need to make sure above Gamma and Lambda operators are evaluated
             * at x(t) and not x(t + dt) if we want to achieve the Forward-Euler scheme
             * for the fluid equations.  Otherwise may get funny behaviors?
             *
             * WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.
             *
             */
            /*
            double tmpPtsX;
            double XX;
            double L = meshDeltaX*numMeshPtsPerDir[d];
            tmpPtsX = SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J]
                    + SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]*deltaT;

            // ensure the position is within the periodic domain
            // because if not LAMMPS does not like it!
            XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
            if (XX > L) {
              XX = XX - floor(XX/L)*L;
            } else if (XX < 0) {
              XX = XX + ceil(-XX/L)*L;
            }
            tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

            SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J] = tmpPtsX;
            */

            // v(t + dt) = v(t) + 0.5*a(t + dt)*dt
            SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]
              += (SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
               / SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->atomMass[I];

           /* NOTE: LAMMPS forces do not have deltaT included in them. */

           if (flagWriteVel) {
             if (d == 0) {
               fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]);
               fflush(fidVel);
             }
           }

          } /* end of d loop */
        } /* end I loop */

      } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

        SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
          = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

        for (I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numControlPts; I++) {
          for (int d = 0; d < num_dim; d++) {

            int J = (I*num_dim) + d;

            /* position x(t + dt) = x(t) + v(t + 0.5*dt)*dt was already updated in initial
             * part of the integration step.
             *
             * The main task here is update the fluid velocity the rest of the half-step.
             *
             * WARNING: Need to make sure above Gamma and Lambda operators are evaluated
             * at x(t) and not x(t + dt) if we want to achieve the Forward-Euler scheme
             * for the fluid equations.  Otherwise may get funny behaviors?
             *
             * WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.
             *
             */
            /*
            double tmpPtsX;
            double XX;
            double L = meshDeltaX*numMeshPtsPerDir[d];
            tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J]
                    + SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]*deltaT;

            // ensure the position is within the periodic domain
            // because if not LAMMPS does not like it!
            XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
            if (XX > L) {
              XX = XX - floor(XX/L)*L;
            } else if (XX < 0) {
              XX = XX + ceil(-XX/L)*L;
            }
            tmpPtsX = XX + (meshCenterX0[d] - L/2.0);

            SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J] = tmpPtsX;
            */

            // v(t + dt) = v(t) + 0.5*a(t + dt)*dt
            SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]
              += (SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J]*0.5*deltaT)
               / SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->atomMass[I];

           /* NOTE: LAMMPS forces do not have deltaT included in them. */

           if (flagWriteVel) {
             if (d == 0) {
               fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]);
               fflush(fidVel);
             }
           }

          } /* end of d loop */
        } /* end I loop */

      } else { /* if not recognized type */
        stringstream message;
        message << "Lagrangian type is not recognized." << endl;
        SELM_Package::packageError(error_str_code, error_str_func, message);
      }

    } /* end of lagrangianI loop */

  } else { /* old way to update the Lagrangian points */

	  for (int couplingOpI = 0;couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {
		  SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

		  if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic) {
			  SELM_Lagrangian *lagrangian = NULL;
			  SELM_CouplingOperator_Uniform1_Periodic *op
				  = (SELM_CouplingOperator_Uniform1_Periodic *)couplingOp;

			  for (int k = 0; k < op->numCoupleList; k++) {
				  lagrangian = op->lagrangianList[k];

				  /* update the lagrangian DOF */
				  if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
					  SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
					  = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;
					  for (I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; I++) {
						  for (int d = 0; d < num_dim; d++) {
							  int J=(I * num_dim) + d;
							  particleForce_dt[couplingOpI][k][J] += SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]*deltaT; /* @@@ ?? what ?? */
							  SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]=particleForce_dt[couplingOpI][k][J];  /* @@@ this will only give last force ?? */
							  double ptsXtemp;
							  ptsXtemp=SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J]
							            + SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J] * deltaT;
							  /* @@@ ?? enforcing periodic boundaries ?? bad idea */
							  /*
							  if (ptsXtemp<lo[d]){
								  ptsXtemp=ptsXtemp+hi[d]-lo[d];
							  } else if (ptsXtemp>hi[d]) {
								  ptsXtemp=ptsXtemp-hi[d]+lo[d];
							  }
							  */
							  SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J]=ptsXtemp;

							  SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]+=particleForce_dt[couplingOpI][k][J]
							          /SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->atomMass[I];
						  }
					  }
				  } else {
					  stringstream message;
					  message << "Lagrangian type is not recognized." << endl;
					  SELM_Package::packageError(error_str_code, error_str_func, message);
				  }
				  delete[] particleForce_dt[couplingOpI][k];
			  }
		  } else {
			  stringstream message;
			  message << "Only Uniform1_Periodic type coupling operator is implemented" << endl;
			  SELM_Package::packageError(error_str_code, error_str_func, message);
		  }
		  delete[] particleForce_dt[couplingOpI];
	  }
	  delete[] particleForce_dt;

	  particleForce_dt = NULL;

  } /* flagOldLagrangianUpdate */

  /* check the velocities of the fluid and particles */
  int flagComputeAvgFluidVelocity = 0;  /* PJA: computes the zero mode of the fluid */

  if ((flagComputeAvgFluidVelocity) && ( (lammps->update->ntimestep - 1) % 1 == 0)) {

    double avgFluidVel[3];
    for (int d = 0; d < num_dim; d++) {
      avgFluidVel[d] = 0.0;
    }

    for(I = 0; I < numMeshPts; I++){
      for (int d = 0; d < num_dim; d++) {
        avgFluidVel[d] += u_main[d][I];
      }
    }

    for (int d = 0; d < num_dim; d++) {
      avgFluidVel[d] = avgFluidVel[d]/numMeshPts;
    }

    double totalFluidMomentum[3];
    double volume = 1.0;
    for (int d = 0; d < num_dim; d++) {
      volume *= numMeshPtsPerDir[d]*meshDeltaX;
    }
    for (int d = 0; d < num_dim; d++) {
      totalFluidMomentum[d] = rho*volume*avgFluidVel[d];
    }

    double ptX[3];
    double ptVel[3];
    double totalPtMomentum[3];
    double totalPtForce[3];

    /* assumes just one Lagrangian type for this test (of known type) */    
    if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {
      SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *lagrangian
        = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)fixSELM->SELM_Lagrangian_List[0];

      for (int d = 0; d < num_dim; d++) {
        ptX[d] = lagrangian->ptsX[d];
      }

      for (int d = 0; d < num_dim; d++) {
        ptVel[d] = lagrangian->pt_Vel[d];
      }

      totalPtMomentum[0] = 0.0; totalPtMomentum[1] = 0.0; totalPtMomentum[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtMomentum[d] += lagrangian->atomMass[I]*lagrangian->pt_Vel[I*num_dim + d];
        }
      }

      totalPtForce[0] = 0.0; totalPtForce[1] = 0.0; totalPtForce[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtForce[d] += lagrangian->pt_Force[I*num_dim + d];
        }
      }
  
    } else if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {
      SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *lagrangian
        = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)fixSELM->SELM_Lagrangian_List[0];

      for (int d = 0; d < num_dim; d++) {
        ptX[d] = lagrangian->ptsX[d];
      }

      for (int d = 0; d < num_dim; d++) {
        ptVel[d] = lagrangian->pt_Vel[d];
      }

      totalPtMomentum[0] = 0.0; totalPtMomentum[1] = 0.0; totalPtMomentum[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtMomentum[d] += lagrangian->atomMass[I]*lagrangian->pt_Vel[I*num_dim + d];
        }
      }

      totalPtForce[0] = 0.0; totalPtForce[1] = 0.0; totalPtForce[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtForce[d] += lagrangian->pt_Force[I*num_dim + d];
        }
      }

    } else if (fixSELM->SELM_Lagrangian_List[0]->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {
      SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *lagrangian
        = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)fixSELM->SELM_Lagrangian_List[0];

      for (int d = 0; d < num_dim; d++) {
        ptX[d] = lagrangian->ptsX[d];
      }

      for (int d = 0; d < num_dim; d++) {
        ptVel[d] = lagrangian->pt_Vel[d];
      }

      totalPtMomentum[0] = 0.0; totalPtMomentum[1] = 0.0; totalPtMomentum[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtMomentum[d] += lagrangian->atomMass[I]*lagrangian->pt_Vel[I*num_dim + d];
        }
      }

      totalPtForce[0] = 0.0; totalPtForce[1] = 0.0; totalPtForce[2] = 0.0;
      for (int I = 0; I < lagrangian->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {
          totalPtForce[d] += lagrangian->pt_Force[I*num_dim + d];
        }
      }

    } else { /* if not recognized type */
      stringstream message;
      message << "Lagrangian type is not recognized." << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

    stringstream message;
    message << "FROM: " << error_str_code << ":" << error_str_func << endl;
    message << "Using Lagrangian type to compute momentum of particles. " << endl;
    message << "  timestep = " << lammps->update->ntimestep << endl;

    for (int d = 0; d < num_dim; d++) {
      message << "  avgFluidVel[" << d << "] = " << avgFluidVel[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  ptX[" << d << "] = " << ptX[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  ptVel[" << d << "] = " << ptVel[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  totalPtForce[" << d << "] = " << totalPtForce[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  totalPtMomentum[" << d << "] = " << totalPtMomentum[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  totalFluidMomentum[" << d << "] = " << totalFluidMomentum[d] << endl;
    }
    for (int d = 0; d < num_dim; d++) {
      message << "  totalSystemMomentum[" << d << "] = " << totalPtMomentum[d] + totalFluidMomentum[d] << endl;
    }

    cout << message.str() << endl;
  }

  // check to see if we should update the fluid next time step
  if (lagCount >= lagNumberTimeSteps - 1) {
    flagUpdateFluid = 1;
    //flagUpdateFluid = 0;
  }

}


void SELM_Integrator_Forward_Euler_Period_Lag::SELM_updateFluidAndStructures_initial()
{
  const char *error_str_func = "SELM_updateFluidAndStructures_initial()";

  /* Velocity Verlet for the particle degrees of freedom is used

      1) v(t + 0.5*dt) = v(t) + 0.5*a(t)*dt
      2) x(t + dt)     = x(t) + v(t + 0.5*dt)dt
      3) compute a(t + dt) from x(t + dt)
      4) v(t + dt)     = v(t + 0.5*dt) + a(t + dt)*dt

      In the initial step we perform 1 and 2 above.  The final performs the steps 3 and 4.
      How to handle the fluid dynamics and update is an interesting question.

      Simple method is to use Forward-Euler which might suffice when lags.

      Verlet-style scheme would be to (1) do Velocity Verlet style update of u to time t + 0.5*dt, (however,
      use v(t + 0.5*dt) throughout, Gamma(t - 0.5*dt)u(t - 0.5*dt), L_u(t - 0.5*dt), F_thm(t - 0.5*dt) etc...).
      (2) perform the rest of the update to u(t + dt) but using the u(t + 0.5*dt) values on RHS so
      we use Gamma(t + 0.5*dt)u(t + 0.5*dt), Lu(t + 0.5*dt), F_thm(t + 0.5*dt), etc...  This scheme is
      analogue of how Langevin Velocity Verlet is implemented in practice.  NOTE: This scheme is
      explicit and avoids any need for solving additional linear systems as in implcit methods.
      Need to check fluctuation-dissipation balance.  Very similar to current Langevin implementations.
      Is this scheme time reversible?  Does it preserve symplectic forms (probably not since dissip.)?
      etc...

      Better approach preserving the time symmetry would be to use Crank-Nicolson
      and better stability would be achieved since this method is unconditionally stable.
      [Note, even if all spatial modes are retained the use of a large time-step would be
       analogous to setting to quasi-steady-state the "fastest" modes...(do analysis of this)]

      Inverting full dissipative operator in practice is non-trivial in terms of the u equations
      since the coupling term contributes to inversion through Lambda and Gamma operators.
      However, if the total momentum description is utilized then these terms do no arise and
      usual Crank-Nicolson can be used.  However, to recover u might impose some additional costs.

      Ideal scheme would be to (1) reformulate in total momentum, (2) do Crank-Nicolson update of p,
      (3) recover u,  In coupling term in particle equations, use 0.5*(Gamma(t) u(t) + Gamma(t+dt) u(t+dt).
      and half time-step velocity V(t + 0.5*dt).

      Another idea is simply to do Crank-Nicolson (semi-implicit) on the Laplacian term and leave the
      drag-coupling operator term explicit at time t.   This would stabalise the scheme with respect to
      Newtonian contributions.


      WARNING!!!! need to implement the fluid update step in a manner consistent with
      the Velocity-Verlet scheme to ensure conservation of total momentum of the system.

      This can be readily done, but need to modify the scheme below!!!!!
      
      WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.


   */

  double  meshDeltaX;
  int     num_dim;
  int    *numMeshPtsPerDir;
  double *meshCenterX0;
  int     numMeshPoints;

  SELM_Eulerian_Uniform1_Periodic *SELM_EulerianData_Uniform1_Periodic = NULL;

  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {

      SELM_EulerianData_Uniform1_Periodic = (SELM_Eulerian_Uniform1_Periodic*) fixSELM->SELM_Eulerian_List[0];

      num_dim     = SELM_EulerianData_Uniform1_Periodic->num_dim;
      meshDeltaX  = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;

      numMeshPtsPerDir = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir;
      meshCenterX0     = SELM_EulerianData_Uniform1_Periodic->meshCenterX0;

      numMeshPoints    = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  } else {
    stringstream message;
    message << "Expecting mesh of type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh was used of type: " << fixSELM->SELM_Eulerian_List[0]->typeStr <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /* -- update the lagrangian DOF */
  //v(t + 0.5*dt) = v(t) + 0.5*a(t)*dt
  //x(t + dt)     = x(t) + v(t + 0.5*dt)dt
  for (int lagrangianI = 0; lagrangianI < fixSELM->SELM_Lagrangian_List_N; lagrangianI++) {
    SELM_Lagrangian *lagrangian = fixSELM->SELM_Lagrangian_List[lagrangianI];

    if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]*deltaT;

          /* REMOVED, since LAMMPS should really handle this and does not like it
           * when we arbitrarily move the points around in our sub-routines...
          XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          if (XX > L) {
            XX = XX - floor(XX/L)*L;
          } else if (XX < 0) {
            XX = XX + ceil(-XX/L)*L;
          }
          tmpPtsX = XX + (meshCenterX0[d] - L/2.0);
          */

          SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

      SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE* SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE
        = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]*deltaT;

          /* REMOVED, since LAMMPS should really handle this and does not like it
           * when we arbitrarily move the points around in our sub-routines...
          XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          if (XX > L) {
            XX = XX - floor(XX/L)*L;
          } else if (XX < 0) {
            XX = XX + ceil(-XX/L)*L;
          }
          tmpPtsX = XX + (meshCenterX0[d] - L/2.0);
          */

          SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

      SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID* SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID
        = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

      for (int I = 0; I < SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->numControlPts; I++) {
        for (int d = 0; d < num_dim; d++) {

          int J = (I*num_dim) + d;

          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]
            += (SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Force[J]*0.5*deltaT)
             / SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->atomMass[I];

          double tmpPtsX;
          double XX;
          double L = meshDeltaX*numMeshPtsPerDir[d];
          tmpPtsX = SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J]
                  + SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]*deltaT;

          /* REMOVED, since LAMMPS should really handle this and does not like it
           * when we arbitrarily move the points around in our sub-routines...
          XX = tmpPtsX - (meshCenterX0[d] - L/2.0);
          if (XX > L) {
            XX = XX - floor(XX/L)*L;
          } else if (XX < 0) {
            XX = XX + ceil(-XX/L)*L;
          }
          tmpPtsX = XX + (meshCenterX0[d] - L/2.0);
          */

          SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->ptsX[J] = tmpPtsX;

         /* NOTE: LAMMPS forces do not have deltaT included in them. */
         if (flagWriteVel) {
           if (d == 0) {
             fprintf(fidVel, "%g ", SELM_LagrangianData_LAMMPS_ATOM_STYLE_ELLIPSOID->pt_Vel[J]);
             fflush(fidVel);
           }
         }

        } /* end of d loop */
      } /* end I loop */

    } else { /* if not recognized type */
      stringstream message;
      message << "Lagrangian type is not recognized." << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }

  } /* end of lagrangianI loop */


}

void SELM_Integrator_Forward_Euler_Period_Lag::syncDomainWithLammpsDomain() {

  const char *error_str_func = "syncDomainWithLammpsDomain()";

  double  meshDeltaX;
  int     num_dim;
  int    *numMeshPtsPerDir;
  double *meshCenterX0;

  SELM_Eulerian_Uniform1_Periodic  *SELM_EulerianData_Uniform1_Periodic;

  /* get data from the integrator */
  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {

    SELM_EulerianData_Uniform1_Periodic
      = (SELM_Eulerian_Uniform1_Periodic *)fixSELM->SELM_Eulerian_List[0];

    num_dim          = SELM_EulerianData_Uniform1_Periodic->num_dim;
    meshDeltaX       = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;
    numMeshPtsPerDir = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir;
    meshCenterX0     = SELM_EulerianData_Uniform1_Periodic->meshCenterX0;

  } else {
    stringstream message;
    message << "Expecting mesh type: %s." << endl;
    message <<   SELM_Eulerian_Types::TYPE_STR_Uniform1_Periodic<< endl;
    message << "Instead mesh type was: " << endl;
    message << fixSELM->SELM_Eulerian_List[0]->typeStr << endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }
}

/* setup for the time step (apply force to the mesh, etc...*/
void SELM_Integrator_Forward_Euler_Period_Lag::start_time_step_Forward_Euler_Period() {

  const char *error_str_func = "start_time_step_Forward_Euler_Period";

  SELM_Eulerian_Uniform1_Periodic   *SELM_EulerianData_Uniform1_Periodic;

  /* WARNING: Need to be careful that location x(t) is used and not x(t + dt) when
   * doing the Forward-Euler scheme for the fluid equations.
   *
   * Velocity Verlet updates x(t) -> x(t + dt) in the initial step so when calling this
   * routine above from final we may get wrong values...
   *
   * However, we would like to use the v(t + 0.5*dt) for the particle velocity when
   * computing the drag forces.  Since this is what is used in the particle equations
   * when computing the drag with Upsilon(v(t + 0.5*dt) - Gamma(t)u(t)).
   *
   */

  int I;
  int i, j, k, d;

  int flagDebugTests = 0;
  int flagDebugStillCompIBForces = 0;

  double *meshCenterX0;
  int    *numMeshPtsPerDir;

  int num_dim;
  int numControlPts;

  //double deltaT;

  double  meshDeltaX;
  double* eulerianV[3];

  int numMeshPoints;

  /* == Dereference the data */
  if (fixSELM->SELM_Eulerian_List[0]->type == SELM_Eulerian_Types::TYPE_Uniform1_Periodic) {

	  SELM_EulerianData_Uniform1_Periodic = (SELM_Eulerian_Uniform1_Periodic*) fixSELM->SELM_Eulerian_List[0];

	  num_dim     = SELM_EulerianData_Uniform1_Periodic->num_dim;
	  meshDeltaX  = SELM_EulerianData_Uniform1_Periodic->meshDeltaX;

	  for(i = 0; i < num_dim; i++) {
	    eulerianV[i] = SELM_EulerianData_Uniform1_Periodic->u_main[i];
	  }

	  numMeshPtsPerDir = SELM_EulerianData_Uniform1_Periodic->numMeshPtsPerDir;

	  numMeshPoints    = numMeshPtsPerDir[0]*numMeshPtsPerDir[1]*numMeshPtsPerDir[2];

  } else {
    stringstream message;
    message << "Expecting mesh of type: " << SELM_Eulerian_Types::TYPE_STR_LAMMPS_SHEAR_UNIFORM1_FFTW3 << endl;
    message << "Instead mesh was used of type: " << fixSELM->SELM_Eulerian_List[0]->typeStr <<  endl;
    SELM_Package::packageError(error_str_code, error_str_func, message);
  }

  /*
  ==============================================================================
  if (flagUpdateFluid)
  ==============================================================================
  */
  if (flagUpdateFluid) {

    /* == Initialize the fluid force density to zero (do this only once) */
    for(d = 0; d < num_dim; d++) {
      fluidForceDensity_lagDt[d] = new double[numMeshPoints];
      for(i = 0; i < numMeshPoints; i++) {
        fluidForceDensity_lagDt[d][i] = 0.0;
      }
    }

    if(flagStochasticDriving) {
  	  computeTimeAvgStochFluct(num_dim, meshDeltaX, numMeshPtsPerDir, lagDeltaT, fixSELM->random, fluidForceDensity_lagDt);
    } // end of if(flagStochasticDriving)

  } /* end flagUpdateFluid */
  /*
  ==============================================================================
  ==============================================================================
  */


  if ((pseudoForceAll_dt == NULL) ||
      (sum_pseudoForceAll_dt == NULL) ||
      (sum_pseudoFluidForce_dt[0] == NULL) || (sum_pseudoFluidForce_dt[1] == NULL) || (sum_pseudoFluidForce_dt[2] == NULL)) {

    // Since only one eulerian grid is implemented and allowed, the Lambda operator can be calculated once
    for (int couplingOpI = 0; couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {

      SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

      if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic) {

        SELM_Lagrangian *lagrangian = NULL;
        SELM_CouplingOperator_Uniform1_Periodic *op
          = (SELM_CouplingOperator_Uniform1_Periodic *)couplingOp;

        for (int k = 0; k < op->numCoupleList; k++) {

          lagrangian = op->lagrangianList[k];

          if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

            SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *lag
              = (SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

            totalPts += lag->numControlPts;

          } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE){

            SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *lag
              = (SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

            totalPts += lag->numControlPts;
            
          } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID){

            SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *lag
              = (SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

            totalPts += lag->numControlPts;

          } else {

            stringstream message;
            message << "Lagrangian Type is not recognized" << endl;
            SELM_Package::packageError(error_str_code, error_str_func, message);

          }

        } /* end k loop */

      } /* end if type == ZZZ */

    } /* end couplingI */

    if (pseudoForceAll_dt == NULL) {
      int N = totalPts*num_dim;
      pseudoForceAll_dt = new double[N];
      for (int I = 0; I < N; I++) {
        pseudoForceAll_dt[I] = 0.0;
      }
    }

    if (sum_pseudoForceAll_dt == NULL) {
      int N = totalPts*num_dim;
      sum_pseudoForceAll_dt = new double[N];
      for (int I = 0; I < N; I++) {
        sum_pseudoForceAll_dt[I] = 0.0;
      }
    }

    if ((sum_pseudoFluidForce_dt[0] == NULL) || (sum_pseudoFluidForce_dt[1] == NULL) || (sum_pseudoFluidForce_dt[2] == NULL)) {
      for (int d = 0; d < num_dim; d++) {
        sum_pseudoFluidForce_dt[d] = new double[numMeshPoints];
        for (int I = 0; I < numMeshPoints; I++) {
          sum_pseudoFluidForce_dt[d][I] = 0.0;
        }
      }
    }

  } /* end pseudoForceAll_dt==NULL or sum_pseudoForceAll_dt==NULL or sum_pseudoFluidForce_dt == NULL NULL */

  double* ptsX;
  ptsX = new double[totalPts*num_dim];

  int globalPtsIndex = 0;

  /* loop over the coupling operators to allow them to compute how Lagrangian based forces are
   * applied to the fluid through the Lambda operators.
   */
  particleForce_dt = new double**[fixSELM->SELM_CouplingOperator_List_N];

  for (int couplingOpI = 0; couplingOpI < fixSELM->SELM_CouplingOperator_List_N; couplingOpI++) {

    /* == Assumes the control point forces were computed by LAMMPS already. */
	SELM_CouplingOperator *couplingOp = fixSELM->SELM_CouplingOperator_List[couplingOpI];

    /* maybe make below more generic */
    if (couplingOp->type == SELM_CouplingOperator_Types::TYPE_Uniform1_Periodic) {

      SELM_Lagrangian *lagrangian = NULL;
      SELM_Eulerian   *eulerian   = NULL;

      SELM_CouplingOperator_Uniform1_Periodic *op
        = (SELM_CouplingOperator_Uniform1_Periodic *)couplingOp;

      particleForce_dt[couplingOpI] = new double*[op->numCoupleList];

      /* loop over the coupled entities */
      for (int k = 0; k < op->numCoupleList; k++) {

        lagrangian = op->lagrangianList[k];
        eulerian   = op->eulerianList[k];

        double* lagVel = NULL;

        /* == Dereference the data */
        if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

          SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *lag
            =(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

          numControlPts = lag->numControlPts;
          lagVel        = lag->pt_Vel;

          int i0, i1;
          for(i = 0; i < numControlPts; i++) {
            i1           = num_dim*globalPtsIndex;
        	  i0           = num_dim*i;

            // new way (want x(t) and not x(t + dt))
        	  // since already updated in initial step
        	  // in Velocity Verlet method
        	  // (undo the update to x)
        	  ptsX[i1]     = lag->ptsX[i0]     - (lag->pt_Vel[i0]*deltaT);
            ptsX[i1 + 1] = lag->ptsX[i0 + 1] - (lag->pt_Vel[i0 + 1]*deltaT);
            ptsX[i1 + 2] = lag->ptsX[i0 + 2] - (lag->pt_Vel[i0 + 2]*deltaT);

            // modify the lagrangian temporarily
            // @@@ WARNING!!! @optimize how this is done
            lag->ptsX[i0]     = ptsX[i1];
            lag->ptsX[i0 + 1] = ptsX[i1 + 1];
            lag->ptsX[i0 + 2] = ptsX[i1 + 2];

        	  // old way
        	  //ptsX[i1]     = lag->ptsX[i0];
        	  //ptsX[i1 + 1] = lag->ptsX[i0 + 1];
        	  //ptsX[i1 + 2] = lag->ptsX[i0 + 2];

        	  globalPtsIndex++;
          } /* end of i loop */

        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

        	SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *lag
        	=(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

        	numControlPts=lag->numControlPts;
        	lagVel=lag->pt_Vel;

        	int i0, i1;
        	for(i=0; i<numControlPts; i++) {
            i1=3*globalPtsIndex;
        		i0=3*i;

            // new way (want x(t) and not x(t + dt))
            // since already updated in initial step
            // in Velocity Verlet method
            // (undo the update to x)
            ptsX[i1]     = lag->ptsX[i0]     - (lag->pt_Vel[i0]*deltaT);
            ptsX[i1 + 1] = lag->ptsX[i0 + 1] - (lag->pt_Vel[i0 + 1]*deltaT);
            ptsX[i1 + 2] = lag->ptsX[i0 + 2] - (lag->pt_Vel[i0 + 2]*deltaT);

            // old way
            //ptsX[i1]=lag->ptsX[i0];
     		    //ptsX[i1+1]=lag->ptsX[i0+1];
    		    //ptsX[i1+2]=lag->ptsX[i0+2];

            // modify the lagrangian temporarily
            // @@@ WARNING!!! @optimize how this is done
            lag->ptsX[i0]     = ptsX[i1];
            lag->ptsX[i0 + 1] = ptsX[i1 + 1];
            lag->ptsX[i0 + 2] = ptsX[i1 + 2];

        		globalPtsIndex++;
        	} /* end of i loop */
        	
        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

        	SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *lag
        	=(SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

        	numControlPts=lag->numControlPts;
        	lagVel=lag->pt_Vel;

        	int i0, i1;
        	for(i=0; i<numControlPts; i++) {
          	i1=3*globalPtsIndex;
        		i0=3*i;

            // new way (want x(t) and not x(t + dt))
            // since already updated in initial step
            // in Velocity Verlet method
            // (undo the update to x)
            ptsX[i1]     = lag->ptsX[i0]     - (lag->pt_Vel[i0]*deltaT);
            ptsX[i1 + 1] = lag->ptsX[i0 + 1] - (lag->pt_Vel[i0 + 1]*deltaT);
            ptsX[i1 + 2] = lag->ptsX[i0 + 2] - (lag->pt_Vel[i0 + 2]*deltaT);

            // old way
            //ptsX[i1]=lag->ptsX[i0];
        		//ptsX[i1+1]=lag->ptsX[i0+1];
        		//ptsX[i1+2]=lag->ptsX[i0+2];

            // modify the lagrangian temporarily
            // @@@ WARNING!!! @optimize how this is done
            lag->ptsX[i0]     = ptsX[i1];
            lag->ptsX[i0 + 1] = ptsX[i1 + 1];
            lag->ptsX[i0 + 2] = ptsX[i1 + 2];

        		globalPtsIndex++;
        	} /* end i loop */

        } else {
          stringstream message;
          message << "Lagrangian Type is not recognized" << endl;
          SELM_Package::packageError(error_str_code, error_str_func, message);
        }

        int numDOFControlPts = num_dim*numControlPts;

        double* pseudoVel                = new double[numDOFControlPts];
        double* pseudoParticleForce      = new double[numDOFControlPts];
        particleForce_dt[couplingOpI][k] = new double[numDOFControlPts];

        for(i = 0; i < numDOFControlPts; i++) {
          particleForce_dt[couplingOpI][k][i] = 0.0;
        }

        for(i = 0; i < numDOFControlPts; i++) {
          pseudoVel[i]           = 0.0;
          pseudoParticleForce[i] = 0.0;
        }

       	//compute pseudo velocity \Gamma u
        // @optimize In lagged code we can simply store the Gamma u between lags since u
        // does not change.
        // WARNING: need to use correct points for this!!! x(t) and not x(t + dt) !!!
       	op->computeOperatorGamma(lagrangian, eulerian, eulerianV, pseudoVel);

       	//compute v - \Gamma u
       	for(i = 0; i < numDOFControlPts; i++) {
       	  pseudoVel[i] = lagVel[i] - pseudoVel[i];
       	}

       	//compute the fluid force acting on the control points PJA: Upsilon*(v - \Gamma u)
       	op->computeOperatorUpsilon(lagrangian, mu, meshDeltaX, pseudoVel, pseudoParticleForce);  /* !!! WARNING: note upsilon needs to be synced with thermal value!!! */
       	for(i = 0; i < numDOFControlPts; i++) {
       	  particleForce_dt[couplingOpI][k][i] = -pseudoParticleForce[i]*deltaT;
       	}

        if(flagStochasticDriving) {

          double* F_thm = new double[numDOFControlPts];
          //double  R_effc;

          //if(op->usePeskin) {
          //  R_effc = 2.0*meshDeltaX;
          //} else {
          //  R_effc = 1.5*meshDeltaX;
          //}

          //double c = 6.0*UNIT_pi*R_effc*mu;
          double c = op->upsilon_0;  /* use value from coupling object (XML) */
          double eta;
          double U;

          int flagRandType = 2;
          if (flagRandType == 1) {  /* use uniform random variates */
            for(i = 0; i < numDOFControlPts; i++) {
              U                                    = fixSELM->random->uniform();
              U                                   -= 0.5; /* center variate */
              F_thm[i]                             = U*sqrt(24.0*KB*T*c*deltaT);
              particleForce_dt[couplingOpI][k][i] += F_thm[i];
            }
          } else if (flagRandType == 2) { /* use gaussian random variates */
            for(i = 0; i < numDOFControlPts; i++) {
              eta                                  = fixSELM->random->gaussian();
              F_thm[i]                             = eta*sqrt(2.0*KB*T*c*deltaT);
              particleForce_dt[couplingOpI][k][i] += F_thm[i];
            }
          } else {
            for(i = 0; i < numDOFControlPts; i++) {
              F_thm[i]                             = 0.0;
              particleForce_dt[couplingOpI][k][i] += F_thm[i];
            }
          }

          delete[] F_thm;
          F_thm = NULL;
        } /* end flagStochasticDriving */

        /* PJA: resets the globalPtsIndex for building later arrays */
        globalPtsIndex -= numControlPts;

       	int i0, i1;
       	for(i = 0; i < numControlPts; i++) {

       	  i0 = num_dim*globalPtsIndex;
       	  i1 = num_dim*i;
       	  for(d = 0; d < num_dim; d++) {
       		  /* PJA: assumes there is only one coupling operator pair per each control point */
       		  pseudoForceAll_dt[i0 + d] = particleForce_dt[couplingOpI][k][i1 + d]; /* @@@ should this be plus += @@@ ??? */
       	  }
       	  globalPtsIndex++;
       	}

        delete[] pseudoVel;
        pseudoVel = NULL;
        delete[] pseudoParticleForce;
        pseudoParticleForce = NULL;

        /* == Dereference the data */
        if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_ANGLE_STYLE) {

          SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *lag
            =(SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE *)lagrangian;

          numControlPts = lag->numControlPts;
          lagVel        = lag->pt_Vel;

          int i0;
          for(i = 0; i < numControlPts; i++) {
            i0           = num_dim*i;

            // new way (want x(t) and not x(t + dt)) for Lambda/Gamma/Upsilon
            // since already updated in initial step
            // in Velocity Verlet method
            // (undo the update to x)
            lag->ptsX[i0]     += lag->pt_Vel[i0]*deltaT;
            lag->ptsX[i0 + 1] += lag->pt_Vel[i0 + 1]*deltaT;
            lag->ptsX[i0 + 2] += lag->pt_Vel[i0 + 2]*deltaT;
          }

        /* end TYPE_LAMMPS_ATOM_ANGLE_STYLE */

        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE) {

          SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *lag
            =(SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE *)lagrangian;

          numControlPts = lag->numControlPts;
          lagVel        = lag->pt_Vel;

          int i0;
          for(i = 0; i < numControlPts; i++) {
            i0           = num_dim*i;

            // new way (want x(t) and not x(t + dt))
            // since already updated in initial step
            // in Velocity Verlet method
            // (undo the update to x)
            lag->ptsX[i0]     += lag->pt_Vel[i0]*deltaT;
            lag->ptsX[i0 + 1] += lag->pt_Vel[i0 + 1]*deltaT;
            lag->ptsX[i0 + 2] += lag->pt_Vel[i0 + 2]*deltaT;
          }

        /* end TYPE_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE */

        } else if (lagrangian->type == SELM_Lagrangian_Types::TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID) {

          SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *lag
            =(SELM_Lagrangian_LAMMPS_ATOM_STYLE_ELLIPSOID *)lagrangian;

          numControlPts = lag->numControlPts;
          lagVel        = lag->pt_Vel;

          int i0;
          for(i = 0; i < numControlPts; i++) {
            i0           = num_dim*i;

            // new way (want x(t) and not x(t + dt))
            // since already updated in initial step
            // in Velocity Verlet method
            // (undo the update to x)
            lag->ptsX[i0]     += lag->pt_Vel[i0]*deltaT;
            lag->ptsX[i0 + 1] += lag->pt_Vel[i0 + 1]*deltaT;
            lag->ptsX[i0 + 2] += lag->pt_Vel[i0 + 2]*deltaT;
          }

          /* end TYPE_LAMMPS_ATOM_STYLE_ELLIPSOID */
        } else { /* if not recognized type */
          stringstream message;
          message << "Lagrangian type is not recognized." << endl;
          SELM_Package::packageError(error_str_code, error_str_func, message);
        } 

      } // end of k loop

    } else {
       stringstream message;
       message << "Only Uniform1_Periodic type coupling operator is implemented" << endl;
       SELM_Package::packageError(error_str_code, error_str_func, message);
	}

  } // end of loop over coupling operators

  // record the current pseudoForce and add to the running sum of
  // forces acting on the fluid over the lagged time period
  int N = totalPts*num_dim;
  for (int I = 0; I < N; I++) {
    sum_pseudoForceAll_dt[I] += pseudoForceAll_dt[I];
  }
  lagCount += 1;

  /* computes Lambda [-(Upsilon*(v - Gamma u))*dT + F_thm*sqrt(dT)] using values stored in pseudoForceAll_dt */
  /* @optimize We can make this much more efficient by simply making the Lambda operator additive which would
   * then just cumulatively sum the results of the various forces acting on the fluid in the fluid density field.
   * WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign.
   */
  SELM_CouplingOperator_Uniform1_Periodic op;  /* WARNING: uses generic coupling op */

  //op.computeOperatorLambda(totalPts, ptsX , SELM_EulerianData_Uniform1_Periodic, pseudoForceAll_dt, pseudoFluidForce_dt);
  //op.computeOperatorLambda(totalPts, ptsX , SELM_EulerianData_Uniform1_Periodic, sum_pseudoForceAll_dt, pseudoFluidForce_dt);

  /* old way of computing the lambda that resets the pseudoFluidForce_dt each calculation
   * and we sum the results later.
   */
  /*
  op.computeOperatorLambda(totalPts, ptsX,
                           SELM_EulerianData_Uniform1_Periodic,
                           pseudoForceAll_dt, pseudoFluidForce_dt);

  for (d = 0; d < num_dim; d++) {
    for (int I = 0; I < numMeshPoints; I++) {
      sum_pseudoFluidForce_dt[d][I] += pseudoFluidForce_dt[d][I];
    }
  }
  */

  // accumulate the force density in the sum_pseudoFluidForce_dt
  // using local operations on the mesh
  // WARNING: make sure pseudoFluidForce_dt is not used elsewhere...
  int flagAdditive = 1;
  op.computeOperatorLambda(totalPts, ptsX,
                           SELM_EulerianData_Uniform1_Periodic,
                           pseudoForceAll_dt, sum_pseudoFluidForce_dt,
                           flagAdditive);

  if (!flagUpdateFluid) {

    /* just record (above) the accumulating pseudoForceAll_dt to be applied later */

  } else { /* update the fluid */
  /*
  ==============================================================================
  if (flagUpdateFluid)
  ==============================================================================
  */
    /* adjust the sign of the fluid force density */
    for(d = 0; d < num_dim; d++) {
      for(i = 0; i < numMeshPoints; i++) {
        /* WARNING: Be sure to take f_thm = g_thm - Lambda[F_thm].  There way typo in SELM 2011 paper, very important to get correlations correct to use sign. */
        fluidForceDensity_lagDt[d][i] -= sum_pseudoFluidForce_dt[d][i]; /* Lambda [(Upsilon*(\bar{v} - Gamma \bar{u}))*dT - \bar{F}_thm*sqrt(dT)] */
      } /* end of i loop */
    } /* end of d loop */

    // reset the pseudoForceAll summation
    int N = totalPts*num_dim;
    for (int I = 0; I < N; I++) {
      sum_pseudoForceAll_dt[I] = 0.0;
    }

    // reset the pseudoFluidForce summation
    for(i = 0; i < numMeshPoints; i++) {
      for(d = 0; d < num_dim; d++) {
        sum_pseudoFluidForce_dt[d][i] = 0.0;
      } /* end of d loop */
    } /* end of i loop */

    /* will reset the update and count once we perform the time step (below) */

  } /* end flagUpdateFluid */
  /*
  ==============================================================================
  ==============================================================================
  */

  //delete[] pseudoForceAll_dt;
  delete[] ptsX;
  ptsX = NULL;

}

void SELM_Integrator_Forward_Euler_Period_Lag::writeSimulationDataToDisk(char *baseFilename,
		                                                                 int timeIndex) {

  const char *error_str_func = "writeSimulationDataToDisk()";

  FILE *fid;
  char  filename[1000];

    /* open the file for writing the data */
    sprintf(filename,"SELM_Integrator%.9d._%s", timeIndex, typeStr);
    fid = fopen(filename,"w");

    if (fid == NULL) {

      stringstream message;
      message << "Could not open file, error occured." << endl;
      message << "  filename = " << filename << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);

    }

    fprintf(fid, "-- SELM_Integrator_Forward_Euler_Period_Lag : Simulation Data -- \n");
    fprintf(fid, "\n");

     /* close the file */
    fclose(fid);

}


void SELM_Integrator_Forward_Euler_Period_Lag::computeLaplacian(double meshDeltaX,
                                                                int *numMeshPtsPerDir,
                                                                double* u_in[3], double* u_out[3]) {

  // computeLaplacian(numMeshPtsPerDir, meshDeltaX, mu, rho, u_main)

  // central difference for laplacian operator on staggered grid
  // for u, v, w
  int k1, k2, k3;
  int k1m1, k1p1;
  int k2m1, k2p1;
  int k3m1, k3p1;

  int I;
  int Im1, I0, Ip1;

  int N1 = numMeshPtsPerDir[0];
  int N2 = numMeshPtsPerDir[1];
  int N3 = numMeshPtsPerDir[2];
  int N12 = N1*N2;

  double lap_u, lap_v, lap_w;
  double dxsq = meshDeltaX * meshDeltaX;

  for (k3 = 0; k3 < N3; k3++) {
    for (k2 = 0; k2 < N2; k2++) {
      for (k1 = 0; k1 < N1; k1++) {

        lap_u = 0.0;
        lap_v = 0.0;
        lap_w = 0.0;

        I = k3*N1*N2 + k2*N1 + k1;

        if (k1 == 0) {
          k1m1 = N1 - 1;
        } else {
          k1m1 = k1 - 1;
        }

        if (k1 == N1 - 1) {
          k1p1 = 0;
        } else {
          k1p1 = k1 + 1;
        }

        if (k2 == 0) {
          k2m1 = N2 - 1;
        } else {
          k2m1 = k2 - 1;
        }

        if (k2 == N2 - 1) {
          k2p1 = 0;
        } else {
          k2p1 = k2 + 1;
        }

        if (k3 == 0) {
          k3m1 = N3 - 1;
        } else {
          k3m1 = k3 - 1;
        }

        if (k3 == N3 - 1) {
          k3p1 = 0;
        } else {
          k3p1 = k3 + 1;
        }

        // x-direction finite difference
        Im1 = k3*N12 + k2*N1 + k1m1;
        I0  = k3*N12 + k2*N1 + k1;
        Ip1 = k3*N12 + k2*N1 + k1p1;

        lap_u += u_in[0][Im1] - 2*u_in[0][I0] + u_in[0][Ip1];

        lap_v += u_in[1][Im1] - 2*u_in[1][I0] + u_in[1][Ip1];

        lap_w += u_in[2][Im1] - 2*u_in[2][I0] + u_in[2][Ip1];

        // y direction finite difference
        Im1 = k3*N12 + k2m1*N1 + k1;
        I0  = k3*N12 +   k2*N1 + k1;
        Ip1 = k3*N12 + k2p1*N1 + k1;

        lap_u += u_in[0][Im1] - 2*u_in[0][I0] + u_in[0][Ip1];

        lap_v += u_in[1][Im1] - 2*u_in[1][I0] + u_in[1][Ip1];

        lap_w += u_in[2][Im1] - 2*u_in[2][I0] + u_in[2][Ip1];

        // z direction finite difference
        Im1 = k3m1*N12 + k2*N1 + k1;
        I0  =   k3*N12 + k2*N1 + k1;
        Ip1 = k3p1*N12 + k2*N1 + k1;

        lap_u += u_in[0][Im1] - 2*u_in[0][I0] + u_in[0][Ip1];

        lap_v += u_in[1][Im1] - 2*u_in[1][I0] + u_in[1][Ip1];

        lap_w += u_in[2][Im1] - 2*u_in[2][I0] + u_in[2][Ip1];

        // divide by 1/dX^2
        lap_u = lap_u / dxsq;
        lap_v = lap_v / dxsq;
        lap_w = lap_w / dxsq;

        /* update the fluid state */
        u_out[0][I] = lap_u;
        u_out[1][I] = lap_v;
        u_out[2][I] = lap_w;

      } /* k1 loop */
    } /* k2 loop */
  } /* k3 loop */

}


void SELM_Integrator_Forward_Euler_Period_Lag::computeTimeAvgStochFluct(int num_dim,
                                                                        double meshDeltaX,
                                                                        int *numMeshPtsPerDir,
                                                                        double effDeltaT,
                                                                         /* gaussDevStateType *gaussDevState, */
                                                                        class RanMars *random,
                                                                        double** fluidStochForceDensity) {
  const char *error_str_func = "computeTimeAvgStochFluct()";

  int d;
  int i, j, k;


  int I;
  int I1, I2;
  int k1, k2, k3;
  int kk1, kk2, kk3;

  double sqrt_d_L_freq[3];
  double d_L_freq;

  fftw_complex* eta_hat[3];

  double meshVolume;
  double cellVolume;


  int N;   //total number of elements for x, y direction
  int N12;
  N12 = numMeshPtsPerDir[0]*numMeshPtsPerDir[1];
  N=N12*numMeshPtsPerDir[2];

  meshVolume = 1.0;
  for (d = 0; d < num_dim; d++) {
    meshVolume = meshVolume * numMeshPtsPerDir[d] * meshDeltaX;
  }
  cellVolume = 1.0;
  for (d = 0; d < num_dim; d++) {
    cellVolume = cellVolume * meshDeltaX;
  }

  eta_hat[0]=(fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
  eta_hat[1]=(fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);
  eta_hat[2]=(fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);

  double sqrtTwo=sqrt(2.0);
  int* indexMatrix[3];
  indexMatrix[0]=new int[N];
  indexMatrix[1]=new int[N];
  indexMatrix[2]=new int[N];

  for (i=0; i<N; i++) {
    indexMatrix[0][i]=0;
    indexMatrix[1][i]=0;
    indexMatrix[2][i]=0;
  }
  for (d=0; d<3; d++) {
    for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
      for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
        for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {
          I = (k3 * N12) + (k2 * numMeshPtsPerDir[0]) + k1;
          if(indexMatrix[d][I]==0){
            kk1 = (numMeshPtsPerDir[0] - k1) % numMeshPtsPerDir[0];
            kk2 = (numMeshPtsPerDir[1] - k2) % numMeshPtsPerDir[1];
            kk3 = (numMeshPtsPerDir[2] - k3) % numMeshPtsPerDir[2];
            I2 = (kk3 * N12) + (kk2 * numMeshPtsPerDir[0]) + kk1;
            if(I==I2) {
                eta_hat[d][I][0]=(double) fixSELM->random->gaussian()/sqrtTwo;
              eta_hat[d][I][1]=0.0;
              indexMatrix[d][I]=1;
            } else {
                eta_hat[d][I][0]=(double) fixSELM->random->gaussian()/sqrtTwo;
              eta_hat[d][I][1]=(double) fixSELM->random->gaussian()/sqrtTwo;
              indexMatrix[d][I]=1;
              eta_hat[d][I2][0]=eta_hat[d][I][0];
              eta_hat[d][I2][1]=-eta_hat[d][I][1];
              indexMatrix[d][I2]=1;
            }
          }
        }
      }
    }
  }

  for (i=0; i<N; i++) {
    if(indexMatrix[0][i]==0||indexMatrix[1][i]==0||indexMatrix[2][i]==0) {
      stringstream message;
      message << "not all random variables are defined." << endl;
      SELM_Package::packageError(error_str_code, error_str_func, message);
    }
  }

  for (i=0; i<num_dim; i++) {
    delete[] indexMatrix[i];
    indexMatrix[i] = NULL;
  }
  //compute u, v, w direction force in Fourier space
  for (k3 = 0; k3 < numMeshPtsPerDir[2]; k3++) {
    for (k2 = 0; k2 < numMeshPtsPerDir[1]; k2++) {
      for (k1 = 0; k1 < numMeshPtsPerDir[0]; k1++) {
            I = (k3 * N12) + (k2 * numMeshPtsPerDir[0]) + k1;
              for (d = 0; d < num_dim; d++) {
                sqrt_d_L_freq[d] = 0.0;
                // uses second order central differences in x and y direction
                d_L_freq = 2.0*(1.0-cos(2.0*UNIT_pi*k1/numMeshPtsPerDir[0]))/(meshDeltaX * meshDeltaX)
                           +2.0*(1.0-cos(2.0*UNIT_pi*k2/numMeshPtsPerDir[1]))/(meshDeltaX * meshDeltaX)
                           +2.0*(1.0-cos(2.0*UNIT_pi*k3/numMeshPtsPerDir[2]))/(meshDeltaX * meshDeltaX);
                sqrt_d_L_freq[d] += d_L_freq;
                if (sqrt_d_L_freq[d] < 0) {
                  stringstream message;
                  message << "sqrt_d_L_freq[" << d << "] = " << sqrt_d_L_freq[d] << " >= 0 failed to hold." << endl;
                  message << "Rough check on the definiteness of the difference operator." << endl;
                  message << "need to be careful of round-off errors here." << endl;
                      message << error_str_code << error_str_func<<endl;
                  SELM_Package::packageError(error_str_code, error_str_func, message);
                }
                sqrt_d_L_freq[d]*=2.0*KB*T*effDeltaT*mu;
                sqrt_d_L_freq[d]=sqrt(sqrt_d_L_freq[d]);
                eta_hat[d][I][0]*=sqrt_d_L_freq[d];
                eta_hat[d][I][1]*=sqrt_d_L_freq[d];
              } // end of d loop
      }
    }
  }

  SELM_Integrator_FFTW3_Period fftwop;
  for(d=0; d<num_dim; d++) {
    fftwop.compute_FFTW3_orthogonal_3dIDFT(numMeshPtsPerDir, eta_hat[d],fluidStochForceDensity[d]);
  }

  //deallocate the pointers
  for(d=0; d<num_dim; d++) {
    fftw_free(eta_hat[d]);
  }
}

void SELM_Integrator_Forward_Euler_Period_Lag::randomGeneratorCheck(class RanMars *random) {
  // check the random number generated is standard normal distribution.
  int i, ii, j, k;
  int N=100;
  double sqrtTwo=sqrt(2.0);
  int* indexMatrix=new int[N];
  fftw_complex* eta_hat;
  const char *error_str_func = "randomGeneratorCheck()";

  eta_hat = (fftw_complex*) fftw_malloc(sizeof(fftw_complex) * N);

  double* g;
  int NN=N*N;

  g=new double[NN];

  for(i=0; i<NN; i++) {
    g[i]=0;
  }

  int totalSimulation=8000;
  double constCoeff=1.0; //-2.0*KB*T*mu*deltaT;
  for(int it=0; it<totalSimulation; it++) {
    for (i=0; i<N; i++) {
      indexMatrix[i]=0;
    }
    for (i = 0; i < N; i++) {
      if(indexMatrix[i]==0){
        ii = (N-i) % N;
        if(i==ii) {
          eta_hat[i][0]=(double) fixSELM->random->gaussian()/sqrtTwo;
          eta_hat[i][1]=0.0;
          indexMatrix[i]=1;
        } else {
          eta_hat[i][0]=(double) fixSELM->random->gaussian()/sqrtTwo;
          eta_hat[i][1]=(double) fixSELM->random->gaussian()/sqrtTwo;
          indexMatrix[i]=1;
          eta_hat[ii][0]=eta_hat[i][0];
          eta_hat[ii][1]=-eta_hat[i][1];
          indexMatrix[i]=1;
        }
      }
    }
    for(i=0; i<N; i++) {
      for(j=0; j<N; j++) {
        k=i*N+j;
        g[k]+=eta_hat[i][0]*eta_hat[j][0]+eta_hat[i][1]*eta_hat[j][1];
      }
    }
    cout<<it<<endl;
  }
  double temp0=0;
  double temp1=0;
  int temp0i, temp1i;
  for(i=0; i<NN; i++) {
    g[i]=g[i]/totalSimulation;
    if(fabs(g[i])>temp0) {
      temp0=fabs(g[i]);
      temp0i=i;
    }
  }
  cout<<temp0i<<"  "<<temp0<<endl;
    stringstream message;
  message << "Random Number Generator Check finished." << endl;
  SELM_Package::packageError(error_str_code, error_str_func, message);
}

}

