/*
 * SELM_Integrator_Forward_Euler_Period_Lag_Lag.h
 *
 * http://atzberger.org/
 *
 */

#ifndef SELM_INTEGRATOR_FORWARD_EULER_PERIOD_LAG_H_
#define SELM_INTEGRATOR_FORWARD_EULER_PERIOD_LAG_H_

#include "SELM_Integrator.h"
#include "SELM_Integrator_Types.h"

#include "SELM_Lagrangian.h"
#include "SELM_Lagrangian_LAMMPS_ATOM_ANGLE_STYLE.h"
#include "SELM_Lagrangian_LAMMPS_HYBRID_CHARGE_ANGLE_STYLE.h"
#include "SELM_Eulerian_Uniform1_Periodic.h"
#include "SELM_Integrator_FFTW3_Period.h"

#include "lammps.h"
#include "fix.h"
#include "domain.h"
#include "irregular.h"

#include "fix_SELM.h"

namespace LAMMPS_NS {

class SELM_Integrator_Forward_Euler_Period_Lag: public LAMMPS_NS::SELM_Integrator {

public:
  // ================= constants =================
  const static int   TYPE;
  const static char* TYPE_STR;

  const static double    UNIT_pi;

  const static char *error_str_code;

  Irregular *irregular;      // for migrating atoms after box flips

  // ================= function prototypes =================
  SELM_Integrator_Forward_Euler_Period_Lag();
  SELM_Integrator_Forward_Euler_Period_Lag(LAMMPS *lmps, FixSELM *fx);
  SELM_Integrator_Forward_Euler_Period_Lag(int, char **);

  void init();
  void setGlobalRefs(LAMMPS *lmps, FixSELM *fx);

  void parse_ParameterFile(char *filename);
  void setup();

  void set_LAMMPS_mask(int *mask_ptr);

  void integrate_initialize();
  void integrate_initial();
  void integrate_final();

  void integrate_initial_VerletTest1();
  void integrate_final_VerletTest1();

  void SELM_updateFluidAndStructures_initial();
  void SELM_updateFluidAndStructures_final();

  void SELM_updateParticlesTest_initial();
  void SELM_updateParticlesTest_final();

  void syncDomainWithLammpsDomain();

  void packageError(int errorCode, void *extras);

  void writeSimulationDataToDisk(char *baseFilename, int timeIndex);

  // ================= variables =================
  int       flagInitializedNumericalMethod;  /*!< Indicates if numerical scheme initialized. */

  int       flagUsePureVerlet;

  int       maxTimeStepIndex;         /*!< Maximum number of time steps to iterate. */
  double    deltaT;                   /*!< Time step for the numerical method. */
  double    mu;
  double    rho;
  double    KB;
  double    T;

  int       flagStochasticDriving;           /*!< Indicates if stochastic forcing is to be used. */
  int       flagIncompressibleFluid;         /*!< Indicates if projection enforcing incompressibility used. */

  int       flagUpdateControlPts;            /*!< Indicates if the numerical method should determine new control
                                                    point locations or if this will be handled elsewhere in the codes. */

  double   *fluidForceDensity_lagDt[3];      // particle force on fluid (summed over lag period duration)
  double ***particleForce_dt;

  int     totalPts;
  double *pseudoForceAll_dt;

  int     lagNumberTimeSteps;
  double  lagDeltaT;
  int     lagCount;
  double *sum_pseudoForceAll_dt;
  double *sum_pseudoFluidForce_dt[3];

  int flagUpdateFluid;

  int  flagWriteVel;
  FILE *fidVel;

private:
  void start_time_step_Forward_Euler_Period();

  void computeTimeAvgStochFluct(int num_dim,
                                double meshDeltaX,
                                int *numMeshPtsPerDir,
                                double effDeltaT,
                                /* gaussDevStateType *gaussDevState, */
                                class RanMars *random,
                                double** fluidStochForceDensity);

  void randomGeneratorCheck(class RanMars *random);

  void computeLaplacian(double meshDeltaX,
                        int *numMeshPtsPerDir,
                        double* u_in[3], double* u_out[3]);

};

}

#endif /* SELM_INTEGRATOR_FORWARD_EULER_PERIOD_LAG_H_ */
