/* ----------------------------------------------------------------------

  Stochastic Eulerian Lagrangian Methods

  Paul J. Atzberger
  http://atzberger.org/
 
------------------------------------------------------------------------- */

#ifdef FIX_CLASS

FixStyle(SELM_statistic_MSD,FixSELM_statistic_MSD)

#else

#ifndef LMP_FIX_SELM_STATISTIC_MSD_H
#define LMP_FIX_SELM_STATISTIC_MSD_H

#ifndef FFT_FFTW /* determine if FFTW package specified */
  #warning "No FFTW package specified for SELM codes.  The SELM functionality will be disabled."
#else
  #define USE_PACKAGE_FFTW3
#endif

#include "stdio.h"
#include "mpi.h"
#include "fix.h"
#include "stddef.h"
#include "random_mars.h"
#include "lammps.h"

#include "Atz_XML_Package.h"

#ifdef USE_PACKAGE_FFTW3
  #include <fftw3.h>
#endif

namespace LAMMPS_NS {

class FixSELM_statistic_MSD : public Fix {

 public:

  static char  *error_str_code;

  /* ================= constants ================= */
  /* constants (values for non-integers defined in .cpp) */
  static const int MAX_STR_LEN = 10000;

  /* parameter file types */
  static const int    PARAM_FILE_TYPE_NULL       = 0;
  static const int    PARAM_FILE_TYPE_TXT        = 1;
  static const int    PARAM_FILE_TYPE_XML        = 2;

  /* ================= function prototypes ================= */
  FixSELM_statistic_MSD(class LAMMPS *, int, char **);

  FixSELM_statistic_MSD(); /* used to construct empty object,
                primarily for testing purposes */

  int          setmask();
  virtual void init();
  void         setup(int);
  void         end_of_step();

  /* =========================== Function Calls =========================== */

  /* write the SELM simulation data to disk */
  void writeAllSimulationData(int timeIndex);

  /* =========================== Variables =========================== */
  LAMMPS *lammps; /* lammps data */

  int      ndelaysteps, nfreq;
  double  *deltaX;
  double  *x_old;

  bigint  nvalid;

  FILE   *fp;

};

}

#endif
#endif
